# python 3.6
# -*- coding: utf-8 -*-
"""
Created on June 11, 17:16:19, 2018
@author: Wade Roberts

For full documentation, see https://waderoberts123.github.io/Hydrostats/
"""
from hydrostats.metrics import *
from hydrostats.analyze import *



# python 3.6
# -*- coding: utf-8 -*-
"""
The visual module contains different plotting functions for time series visualization. It allows
users to plot hydrographs, scatter plots, histograms, and quantile-quantile (qq) plots to visualize
time series data. In some of the visualization functions, metrics can be added to the plots for a
more complete summary of the data.
"""
from __future__ import division
from hydrostats.metrics import function_list, metric_abbr
from HydroErr.HydroErr import treat_values
import numpy as np
import matplotlib.pyplot as plt
import calendar
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

__all__ = ['plot', 'hist', 'scatter', 'qqplot']


def plot(merged_data_df, legend=('Simulated Data', 'Observed Data'), metrics=None, grid=False, title=None,
         x_season=False, labels=None, linestyles=('ro', 'b^'), tight_xlim=False, fig_size=(10, 6),
         text_adjust=(-0.35, 0.75), plot_adjust=0.27, transparency=0.5, ebars=None, ecolor=None,
         markersize=2, errorevery=1, markevery=1):
    """
    Create a comparison time series line plot of simulated and observed time series data.

    The time series plot is a function that is available for viewing two times series plotted side
    by side vs. time. Goodness of fit metrics can also be viewed on the plot to compare the two
    time series.

    Parameters
    ----------
    merged_data_df: DataFrame
        DataFrame must contain datetime index and floating point type numbers in the two columns.
        The left columns must be simulated data and the right column observed data.

    legend: tuple of str
        Adds a Legend in the 'best' location determined by matplotlib. The entries in the tuple describe the left and
        right columns of the merged_data_df data.

    metrics: list of str
        Adds Metrics to the left side of the plot. Any metric from the Hydrostats library can
        be added to the plot as the abbreviation of the function. The entries must be in a list.
        (e.g. ['ME', 'r2', 'KGE (2012)']).

    grid: bool
        If True, adds a grid to the plot.

    title: str
        If given, adds a title to the plot.

    x_season: bool
        If True, the x-axis ticks will be monthly. This is a useful feature when plotting seasonal
        time series comparisons (e.g. daily averages).

    labels: list of str
        List of two str type inputs specifying x-axis labels and y-axis labels, respectively.

    linestyles: list of str
        List of two string type inputs thet will change the linestyle of the predicted and
        recorded data, respectively. Linestyle references can be found in
        :ref:`matplotlib_linestyles`.

    tight_xlim: bool
        If true, will set the padding to zero for the lines in the line plot.

    fig_size: tuple of floats
        Tuple of length two that specifies the horizontal and vertical lengths of the plot in
        inches, respectively.

    text_adjust: tuple
        Tuple of length two with float type inputs indicating the relative position of the text
        (x-coordinate, y-coordinate) when adding metrics to the plot.

    plot_adjust: float
        Specifies the relative position to shift the plot the the right when adding metrics to the
        plot.

    transparency: float
        Value between 0 to 1 indicating the transparency of the two lines that are plotted and
        error bars if they are plotted, lower means more transparent.

    ebars: DataFrame
        DataFrame must contain datetime index and two columns of data that specify the error
        of the plots, with the simulated error on the left and the observed error on the right.
        These dataframes can be created with the
        :doc:`daily_std_error <hydrostats.data.daily_std_error>`,
        :doc:`daily_std_dev <hydrostats.data.daily_std_dev>`,
        :doc:`monthly_std_error <hydrostats.data.monthly_std_error>`, and
        :doc:`monthly_std_dev <hydrostats.data.monthly_std_dev>` functions.

    ecolor: tuple of str
        Tuple of two sting type inputs specifying the colors of the errorbars (e.g. ['r', 'k'] would
        make red and black errorbars on the simulated and observed data, respectively).

    markersize: float
        Indicates the size of the markers on the plot, if markers are used.

    errorevery: int
        Specifies how often to put error bars on the plot.

    markevery: int
        Specifies how often to put markers on the plot if markers are used.

    Returns
    -------
    fig : Matplotlib figure instance
        A matplotlib figure handle is returned, which can be viewed with the matplotlib.pyplot.show() command.

    Examples
    --------

    In this example two models are compared.

    >>> import hydrostats.data as hd
    >>> import hydrostats.visual as hv
    >>> import matplotlib.pyplot as plt

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'

    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))
    >>> seasonal_df = hd.seasonal_period(merged_df, ('04-01', '07-31'), time_range=('1986-01-01', '1992-12-31'))
    >>> daily_avg_df = hd.daily_average(merged_data=merged_df)  # Seasonal Daily Averages
    >>> daily_std_error = hd.daily_std_error(merged_data=merged_df)  # Seasonal Daily Standard Deviation

    The entire timeseries is plotted below

    >>> plot(merged_data_df=merged_df,
    >>>      title='Hydrograph of Entire Time Series',
    >>>      linestyles=['r-', 'k-'],
    >>>      legend=('SFPT', 'GLOFAS'),
    >>>      labels=['Datetime', 'Streamflow (cfs)'],
    >>>      metrics=['ME', 'NSE', 'SA'],
    >>>      grid=True)
    >>> plt.show()

    .. image:: /Figures/plot_full1.png

    The seasonal averages with standard error bars is plotted below

    >>> plot(merged_data_df=daily_avg_df,
    >>>      title='Daily Average Streamflow (Standard Error)',
    >>>      legend=('SFPT', 'GLOFAS'),
    >>>      x_season=True,
    >>>      labels=['Datetime', 'Streamflow (csm)'],
    >>>      linestyles=['r-', 'k-'],
    >>>      fig_size=(14, 8),
    >>>      ebars=daily_std_error,
    >>>      ecolor=('r', 'k'),
    >>>      tight_xlim=True
    >>>      )
    >>> plt.show()

    .. image:: /Figures/plot_seasonal.png

    """
    fig = plt.figure(figsize=fig_size, facecolor='w', edgecolor='k')
    ax = fig.add_subplot(111)

    # Setting Variable for the simulated data, observed data, and time stamps
    sim = merged_data_df.iloc[:, 0].values
    obs = merged_data_df.iloc[:, 1].values
    time = merged_data_df.index.values

    # Plotting the Data
    if ebars is None:
        plt.plot(time, sim, linestyles[0], markersize=markersize,
                 label=legend[0], alpha=transparency, markevery=markevery)
        plt.plot(time, obs, linestyles[1], markersize=markersize,
                 label=legend[1], alpha=transparency, markevery=markevery)
        plt.legend(fontsize=14)
    elif ebars is not None:
        plt.errorbar(x=time, y=sim, yerr=ebars.iloc[:, 0].values,
                     fmt=linestyles[0], markersize=markersize, label=legend[0], alpha=transparency, ecolor=ecolor[0],
                     markevery=markevery, errorevery=errorevery)
        plt.errorbar(x=time, y=obs, yerr=ebars.iloc[:, 1].values,
                     fmt=linestyles[1], markersize=markersize, label=legend[1], alpha=transparency, ecolor=ecolor[1],
                     markevery=markevery, errorevery=errorevery)
        plt.legend(fontsize=14)

    # Adjusting the plot if user wants tight x axis limits
    if tight_xlim:
        plt.xlim(time[0], time[-1])

    # Changing the X axis for a better seasonal plot if seasonal and adjusting tick sizes
    if x_season:
        seasons = calendar.month_abbr[1:13]
        day_month = np.array([31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
        day_month_cum = np.cumsum(day_month)
        fractions = day_month_cum[:11] / 366
        fractions = np.insert(fractions, 0, 0)
        index = np.rint(fractions * len(time)).astype(np.integer)
        plt.xticks(time[index], seasons, fontsize=14, rotation=45)
    else:
        plt.xticks(fontsize=14, rotation=45)

    plt.yticks(fontsize=14)

    # Placing Labels if requested
    if labels:
        # Plotting Labels
        plt.xlabel(labels[0], fontsize=18)
        plt.ylabel(labels[1], fontsize=18)
    if title:
        title_dict = {'family': 'sans-serif',
                      'color': 'black',
                      'weight': 'normal',
                      'size': 20,
                      }
        ax.set_title(label=title, fontdict=title_dict, pad=25)

    # Placing a grid if requested
    if grid:
        plt.grid(True)

    # Fixes issues with parts of plot being cut off
    plt.tight_layout()

    # Placing Metrics on the Plot if requested
    if metrics:
        function_list_str = metric_abbr

        assert isinstance(metrics, list)

        for metric in metrics:
            assert metric in function_list_str

        index = []
        for metric in metrics:
            index.append(function_list_str.index(metric))

        selected_metrics = []
        for i in index:
            selected_metrics.append(
                function_list_str[i] + '=' + str(round(function_list[i](sim, obs), 3)))

        formatted_selected_metrics = ''
        for i in selected_metrics:
            formatted_selected_metrics += i + '\n'

        font = {'family': 'sans-serif',
                'weight': 'normal',
                'size': 14}
        plt.text(text_adjust[0], text_adjust[1], formatted_selected_metrics, ha='left', va='center',
                 transform=ax.transAxes, fontdict=font)

        plt.subplots_adjust(left=plot_adjust)

    return fig


def hist(merged_data_df=None, sim_array=None, obs_array=None, num_bins=100, z_norm=False,
         legend=('Simulated', 'Observed'), grid=False, title=None, labels=None, prob_dens=False,
         figsize=(12, 6)):
    """Plots a histogram comparing simulated and observed data.

    The histogram plot is a function that is available for comparing the histograms of two time
    series. Data can be Z-score normalized as well as fit in a probability density function.

    Parameters
    ----------
    merged_data_df: DataFrame
        Dataframe must contain a datetime type index and floating point type numbers in two
        columns. The left column must be simulated data and the right column must be observed data.
        If given, sim_array and obs_array must be None.

    sim_array: 1D ndarray
        Array of simulated data. If given, merged_data_df parameter must be None and obs_array must
        be given.

    obs_array: 1D ndarray
        Array of observed data. If given, merged_data_df parameter must be None and sim_array must
        be given.

    num_bins: int
        Specifies the number of bins in the histogram.

    z_norm: bool
        If True, the data will be Z-score normalized.

    legend: tuple of str
        Tuple of length two with str inputs. Adds a Legend in the 'best' location determined by
        matplotlib. The entries in the tuple label the simulated and observed data
        (e.g. ['Simulated Data', 'Predicted Data']).

    grid: bool
        If True, adds a grid to the plot.

    title: str
        If given, sets the title of the plot.

    labels: tuple of str
        Tuple of two string type objects to set the x-axis labels and y-axis labels, respectively.

    prob_dens: bool
        If True, normalizes both histograms to form a probability density, i.e., the area
        (or integral) under each histogram will sum to 1.

    figsize: tuple of float
        Tuple of length two that specifies the horizontal and vertical lengths of the plot in
        inches, respectively.

    Returns
    -------
    fig : Matplotlib figure instance
        A matplotlib figure handle is returned, which can be viewed with the matplotlib.pyplot.show() command.

    Examples
    --------

    In this example the histograms of two models are compared to check their distributions

    >>> import hydrostats.data as hd
    >>> import hydrostats.visual as hv
    >>> import matplotlib.pyplot as plt

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))

    The histogram with 100 bins is plotted below

    >>> hist(merged_data_df=merged_df,
    >>>      num_bins=100,
    >>>      title='Histogram of Streamflows',
    >>>      legend=('SFPT', 'GLOFAS'),
    >>>      labels=('Bins', 'Frequency'),
    >>>      grid=True)
    >>> plt.show()

    .. image:: /Figures/hist1.png

    """
    # Getting the fig and axis handles
    fig, ax1 = plt.subplots(figsize=figsize)

    if merged_data_df is not None and sim_array is None and obs_array is None:
        # Creating a simulated and observed data array
        sim = merged_data_df.iloc[:, 0].values
        obs = merged_data_df.iloc[:, 1].values
    elif sim_array is not None and obs_array is not None and merged_data_df is None:
        sim = sim_array
        obs = obs_array
    else:
        raise RuntimeError("You must either pass in a dataframe or two arrays.")

    if z_norm:
        # Calculating the Z Scores for the simulated data
        sim_mean = np.mean(sim)
        sim_std_dev = np.std(sim)
        # The z scores override sim from before because we are plotting the Z scores
        sim = ((sim - sim_mean) / sim_std_dev)

        # Calculating the Z Scores for the observed data
        obs_mean = np.mean(obs)
        obs_std_dev = np.std(obs)
        # The z scores override obs from before because we are plotting the Z scores
        obs = ((obs - obs_mean) / obs_std_dev)

        # Finding the maximum and minimum Z scores
        sim_max = np.max(sim)
        sim_min = np.min(sim)
        obs_max = np.max(obs)
        obs_min = np.min(obs)

        total_max = np.max([sim_max, obs_max])
        total_min = np.min([sim_min, obs_min])

        # Creating the bins based on the max and min
        bins = np.linspace(total_min - 0.01, total_max + 0.01, num_bins)
    else:
        # Calculating the max and min of both data sets
        sim_max = np.max(sim)
        sim_min = np.min(sim)
        obs_max = np.max(obs)
        obs_min = np.min(obs)

        total_max = np.max([sim_max, obs_max])
        total_min = np.min([sim_min, obs_min])

        # Creating the bins based on the max and min
        bins = np.linspace(total_min - 0.01, total_max + 0.01, num_bins)

    if legend is None:
        # Plotting the data without the legend
        ax1.hist(sim, bins, alpha=0.5, edgecolor='black', linewidth=0.5, density=prob_dens)
        ax1.hist(obs, bins, alpha=0.5, edgecolor='black', linewidth=0.5, density=prob_dens)
    else:
        # Plotting the data with the legend
        ax1.hist(sim, bins, alpha=0.5, label=legend[0], edgecolor='black', linewidth=0.5, density=prob_dens)
        ax1.hist(obs, bins, alpha=0.5, label=legend[1], edgecolor='black', linewidth=0.5, density=prob_dens)
        ax1.legend(framealpha=1)

    # Setting the x and y tick size
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    # Creating a grid
    if grid:
        plt.grid(True)

    # Creating a title
    if title:
        title_dict = {'family': 'sans-serif',
                      'color': 'black',
                      'weight': 'normal',
                      'size': 20,
                      }
        ax1.set_title(label=title, fontdict=title_dict, pad=15)

    # Creating x and y axis labels
    if labels:
        plt.xlabel(labels[0], fontsize=18)
        plt.ylabel(labels[1], fontsize=18)

    # Assuring a tight layout
    plt.tight_layout()

    return fig


def scatter(merged_data_df=None, sim_array=None, obs_array=None, grid=False, title=None, labels=None, best_fit=False,
            marker_style='ko', metrics=None, log_scale=False, line45=False, figsize=(12, 8)):
    """Creates a scatter plot of the observed and simulated data.

    Parameters
    ----------
    merged_data_df: DataFrame
        Dataframe must contain a datetime type index and floating point type numbers in two
        columns. The left column must be simulated data and the right column must be observed data.
        If given, sim_array and obs_array must be None.

    sim_array: 1D ndarray
        Array of simulated data. If given, merged_data_df parameter must be None and obs_array must
        be given.

    obs_array: 1D ndarray
        Array of observed data. If given, merged_data_df parameter must be None and sim_array must
        be given.

    grid: bool
        If True, adds a grid to the plot.

    title: str
        If given, sets the title of the plot.

    labels: tuple of str
        Tuple of two string type objects to set the x-axis labels and y-axis labels, respectively.

    best_fit: bool
        If True, adds a best linear regression line on the graph
        with the equation for the line in the legend.

    marker_style: str
        If give, changes the markerstyle of the points on the scatter plot. Matplotlib styling
        guides are found in :ref:`matplotlib_linestyles`.

    metrics: list of str
        Adds Metrics to the left side of the plot. Any metric from the Hydrostats library can
        be added to the plot as the abbreviation of the function. The entries must be in a list.
        (e.g. ['ME', 'r2', 'KGE (2012)']).

    log_scale: bool
        If True, log-log scale will be used on the scatter plot.

    line45: bool
        IF Trre, adds a 45 degree line to the plot and the legend.

    figsize: tuple of float
        Tuple of length two that specifies the horizontal and vertical lengths of the plot in
        inches, respectively.

    Returns
    -------
    fig : Matplotlib figure instance
        A matplotlib figure handle is returned, which can be viewed with the matplotlib.pyplot.show() command.


    Examples
    --------

    A scatter plot is created in this example comparing two models.

    >>> import hydrostats.data as hd
    >>> import hydrostats.visual as hv
    >>> import matplotlib.pyplot as plt

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))

    >>> sim_array = merged_df.iloc[:, 0].values
    >>> obs_array = merged_df.iloc[:, 1].values

    >>> scatter(merged_data_df=merged_df, grid=True, title='Scatter Plot (Normal Scale)',
    >>>         labels=('SFPT', 'GLOFAS'), best_fit=True)
    >>> plt.show()

    .. image:: /Figures/scatter.png

    Arrays can be used as well in the parameters, as demonstrated below.

    >>> scatter(sim_array=sim_array, obs_array=obs_array, grid=True, title='Scatter Plot (Log-Log Scale)',
    >>>         labels=('SFPT', 'GLOFAS'), line45=True, metrics=['ME', 'KGE (2012)'])
    >>> plt.show()

    .. image:: /Figures/scatterlog.png

    """
    fig = plt.figure(figsize=figsize, facecolor='w', edgecolor='k')
    ax = fig.add_subplot(111)

    if merged_data_df is not None and sim_array is None and obs_array is None:
        # Creating a simulated and observed data array
        sim = merged_data_df.iloc[:, 0].values
        obs = merged_data_df.iloc[:, 1].values
    elif sim_array is not None and obs_array is not None and merged_data_df is None:
        sim = sim_array
        obs = obs_array
    else:
        raise RuntimeError("You must either pass in a dataframe or two arrays.")

    max_both = max([np.max(sim), np.max(obs)])

    if not log_scale:
        plt.plot(sim, obs, marker_style)
    else:
        plt.loglog(sim, obs, marker_style)

    if line45:
        plt.plot(np.arange(0, int(max_both) + 1), np.arange(0, int(max_both) + 1), 'r--', label='45$^\circ$ Line')

    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    if grid:
        plt.grid(True)

    if title:
        plt.title(title, fontsize=20)

    if labels:
        plt.xlabel(labels[0], fontsize=18)
        plt.ylabel(labels[1], fontsize=18)

    if best_fit:
        # Getting a polynomial fit and defining a function with it
        p = np.polyfit(sim, obs, 1)
        f = np.poly1d(p)

        # Calculating new x's and y's
        x_new = np.linspace(0, sim.max(), sim.size)
        y_new = f(x_new)

        # Formatting the best fit equation to be able to display in latex
        equation = "{} x + {}".format(np.round(p[0], 4), np.round(p[1], 4))

        # Plotting the best fit line with the equation as a legend in latex
        plt.plot(x_new, y_new, 'k', label="${}$".format(equation))

    if line45 or best_fit:
        plt.legend(fontsize=12)

    if metrics is not None:
        function_list_str = metric_abbr

        assert isinstance(metrics, list)
        for metric in metrics:
            assert metric in function_list_str
        index = []
        for metric in metrics:
            index.append(function_list_str.index(metric))
        selected_metrics = []
        for i in index:
            selected_metrics.append(function_list_str[i] + '=' +
                                    str(round(function_list[i](sim, obs), 3)))
        formatted_selected_metrics = ''
        for i in selected_metrics:
            formatted_selected_metrics += i + '\n'
        font = {'family': 'sans-serif',
                'weight': 'normal',
                'size': 14}
        plt.text(-0.35, 0.75, formatted_selected_metrics, ha='left', va='center', transform=ax.transAxes, fontdict=font)
        plt.subplots_adjust(left=0.25)

    return fig


def qqplot(merged_data_df=None, sim_array=None, obs_array=None, interpolate='linear', title=None,
           xlabel='Simulated Data Quantiles', ylabel='Observed Data Quantiles', legend=False, replace_nan=None,
           replace_inf=None, remove_neg=False, remove_zero=False, figsize=(12, 8)):
    """Plots a Quantile-Quantile plot of the simulated and observed data.

    Useful for comparing to see whether the two datasets come from the same distribution.

    Parameters
    ----------
    merged_data_df: DataFrame
        Dataframe must contain a datetime type index and floating point type numbers in two
        columns. The left column must be simulated data and the right column must be observed data.
        If given, sim_array and obs_array must be None.

    sim_array: 1D ndarray
        Array of simulated data. If given, merged_data_df parameter must be None and obs_array must
        be given.

    obs_array: 1D ndarray
        Array of observed data. If given, merged_data_df parameter must be None and sim_array must
        be given.

    interpolate: str
        Specifies the interpolation type when computing quantiles.

    title: str
        If given, sets the title of the plot.

    xlabel: str
        The label for the x axis that holds the simulated data quantiles.

    ylabel: str
        The label for the y axis that holds the observed data quantiles.

    legend: bool
        If True, a legend to explain the elements on the plot will be added.

    replace_nan: float, optional
        If given, indicates which value to replace NaN values with in the two arrays. If None, when
        a NaN value is found at the i-th position in the observed OR simulated array, the i-th value
        of the observed and simulated array are removed before the computation.

    replace_inf: float, optional
        If given, indicates which value to replace Inf values with in the two arrays. If None, when
        an inf value is found at the i-th position in the observed OR simulated array, the i-th
        value of the observed and simulated array are removed before the computation.

    remove_neg: boolean, optional
        If True, when a negative value is found at the i-th position in the observed OR simulated
        array, the i-th value of the observed AND simulated array are removed before the
        computation.

    remove_zero: boolean, optional
        If true, when a zero value is found at the i-th position in the observed OR simulated
        array, the i-th value of the observed AND simulated array are removed before the
        computation.

    figsize: tuple of float
        Tuple of length two that specifies the horizontal and vertical lengths of the plot in
        inches, respectively.

    Returns
    -------
    fig : Matplotlib figure instance
        A matplotlib figure handle is returned, which can be viewed with the matplotlib.pyplot.show() command.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import hydrostats.visual as hv
    >>> import matplotlib.pyplot as plt

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))

    >>> qqplot(merged_data_df=merged_df, title='Quantile-Quantile Plot of Data',
    >>>        xlabel='SFPT Data Quantiles', ylabel='GLOFAS Data Quantiles', legend=True,
    >>>        figsize=(8, 6))
    >>> plt.show()

    .. image:: /Figures/qqplot.png

    """

    fig = plt.figure(figsize=figsize, facecolor='w', edgecolor='k')

    if merged_data_df is not None and sim_array is None and obs_array is None:
        # Creating a simulated and observed data array
        sim = merged_data_df.iloc[:, 0].values
        obs = merged_data_df.iloc[:, 1].values
    elif sim_array is not None and obs_array is not None and merged_data_df is None:
        sim = sim_array
        obs = obs_array
    else:
        raise RuntimeError("You must either pass in a dataframe or two arrays.")

    sim, obs = treat_values(sim, obs, replace_nan=replace_nan, replace_inf=replace_inf, remove_neg=remove_neg,
                            remove_zero=remove_zero)

    # Finding the size of n and creating a percentile vector:
    n = sim.size

    pvec = 100 * ((np.arange(1, n + 1) - 0.5) / n)

    sim_perc = np.percentile(sim, pvec, interpolation=interpolate)
    obs_perc = np.percentile(obs, pvec, interpolation=interpolate)

    # Finding the interquartile range to plot the best fit line
    quant_1_sim = np.percentile(sim, 25, interpolation=interpolate)
    quant_3_sim = np.percentile(sim, 75, interpolation=interpolate)
    quant_1_obs = np.percentile(obs, 25, interpolation=interpolate)
    quant_3_obs = np.percentile(obs, 75, interpolation=interpolate)
    quant_sim = np.array([quant_1_sim, quant_3_sim])
    quant_obs = np.array([quant_1_obs, quant_3_obs])

    dsim = quant_3_sim - quant_1_sim
    dobs = quant_3_obs - quant_1_obs
    slope = dobs / dsim
    centersim = (quant_1_sim + quant_3_sim) / 2
    centerobs = (quant_1_obs + quant_3_obs) / 2
    maxsim = np.max(sim)
    minsim = np.min(sim)
    maxobs = centerobs + slope * (maxsim - centersim)
    minobs = centerobs - slope * (centersim - minsim)

    msim = np.array([minsim, maxsim])
    mobs = np.array([minobs, maxobs])

    if not legend:
        plt.plot(sim_perc, obs_perc, 'b^', markersize=2)
        plt.plot(msim, mobs, 'r-.', lw=1)
        plt.plot(quant_sim, quant_obs, 'r-', marker='o', mfc='k', lw=2)
    else:
        plt.plot(sim_perc, obs_perc, 'b^', markersize=2, label='Quantiles')
        plt.plot(msim, mobs, 'r-.', lw=1, label='Entire Range of Quantiles')
        plt.plot(quant_sim, quant_obs, 'r-', marker='o', mfc='w', lw=2, label='Inter-Quartile Range')
        plt.legend(fontsize=14)

    if title is not None:
        plt.title(title, fontsize=18)

    # Formatting things
    plt.xlabel(xlabel, fontsize=16)
    plt.ylabel(ylabel, fontsize=16)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.tight_layout()

    return fig


if __name__ == "__main__":
    pass

    # import pandas as pd
    #
    # merged_df = pd.read_pickle("tests/Files_for_tests/merged_df.pkl")
    #
    # plot(merged_data_df=merged_df,
    #      title='Hydrograph of Entire Time Series',
    #      linestyles=['r-', 'k-'],
    #      legend=('SFPT', 'GLOFAS'),
    #      labels=['Datetime', 'Streamflow (cfs)'],
    #      metrics=['ME', 'NSE', 'SA'],
    #      grid=True)
    #
    # plt.savefig(r"tests/baseline_images/plot_tests/plot_full1_test.png")
    #
    # plot(merged_data_df=daily_avg_df,
    #         title='Daily Average Streamflow (Standard Error)',
    #         legend=('SFPT', 'GLOFAS'),
    #         x_season=True,
    #         labels=['Datetime', 'Streamflow (csm)'],
    #         linestyles=['r-', 'k-'],
    #         fig_size=(14, 8),
    #         ebars=daily_std_error,
    #         ecolor=('r', 'k'),
    #         tight_xlim=True
    #         )



# python 3.6
# -*- coding: utf-8 -*-
"""
The ens_metrics module contains all of the metrics included in hydrostats that measure forecast
skill. Each forecast metric is contained in a function, and every metric has the ability to treat
missing values as well as remove zero and negative values from the timeseries data. Users will be
warned which start dates have been removed in the warnings that display during the function
execution.
"""
from __future__ import division
from hydrostats.metrics import pearson_r
import numpy as np
from numba import jit, prange
import warnings

__all__ = ["ens_me", "ens_mae", "ens_mse", "ens_rmse", "ens_pearson_r", "crps_hersbach",
           "crps_kernel", "ens_crps", "ens_brier", "auroc", "skill_score"]


def ens_me(obs, fcst_ens=None, remove_zero=False, remove_neg=False):
    """Calculate the mean error between observed values and the ensemble mean.

    Parameters
    ----------

    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    Returns
    -------
    float
        The mean error between the observed time series data and the ensemble mean.

    Examples
    --------
    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array

    >>> ensemble_array = (np.random.rand(100, 52) + 1) * 100  # 52 Ensembles
    >>> observed_array = (np.random.rand(100) + 1) * 100

    Computing the ME between the ensemble mean and the observed data. Note that because the data is
    random the errors cancel out, leaving a low ME value.

    >>> em.ens_me(obs=observed_array, fcst_ens=ensemble_array)
    -2.5217349574908074

    """
    # Treating data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    fcst_ens_mean = np.mean(fcst_ens, axis=1)

    error = fcst_ens_mean - obs
    return np.mean(error)


def ens_mae(obs, fcst_ens=None, remove_zero=False, remove_neg=False):
    """Calculate the mean absolute error between observed values and the ensemble mean.

    Parameters
    ----------
    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    Returns
    -------
    float
        The mean absolute error between the observed time series data and the ensemble mean.

    Examples
    --------
    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array

    >>> ensemble_array = (np.random.rand(100, 52) + 1) * 100  # 52 Ensembles
    >>> observed_array = (np.random.rand(100) + 1) * 100

    Computing the ME between the ensemble mean and the observed data. Note that because the data is
    random the errors cancel out, leaving a low ME value.

    >>> em.ens_mae(obs=observed_array, fcst_ens=ensemble_array)
    26.35428724003365

    """
    # Treating data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    fcst_ens_mean = np.mean(fcst_ens, axis=1)

    error = fcst_ens_mean - obs
    return np.mean(np.abs(error))


def ens_mse(obs, fcst_ens=None, remove_zero=False, remove_neg=False):
    """Calculate the mean squared error between observed values and the ensemble mean.

    Parameters
    ----------
    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    Returns
    -------
    float
        The mean error between the observed time series data and the ensemble mean.

    Examples
    --------
    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array

    >>> ensemble_array = (np.random.rand(100, 52) + 1) * 100  # 52 Ensembles
    >>> observed_array = (np.random.rand(100) + 1) * 100

    Computing the MSE between the ensemble mean and the observed data

    >>> em.ens_mse(obs=observed_array, fcst_ens=ensemble_array)
    910.5648405687582

    """
    # Treating data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    fcst_ens_mean = np.mean(fcst_ens, axis=1)

    error = fcst_ens_mean - obs
    return np.mean(error ** 2)


def ens_rmse(obs, fcst_ens=None, remove_zero=False, remove_neg=False):
    """Calculate the root mean squared error between observed values and the ensemble mean.

    Parameters
    ----------
    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    Returns
    -------
    float
        The mean error between the observed time series data and the ensemble mean.

    Examples
    --------
    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array

    >>> ensemble_array = (np.random.rand(100, 52) + 1) * 100  # 52 Ensembles
    >>> observed_array = (np.random.rand(100) + 1) * 100

    Computing the MSE between the ensemble mean and the observed data

    >>> em.ens_rmse(obs=observed_array, fcst_ens=ensemble_array)
    30.17556694693172

    """
    # Treating data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    fcst_ens_mean = np.mean(fcst_ens, axis=1)

    error = fcst_ens_mean - obs
    return np.sqrt(np.mean(error ** 2))


def ens_pearson_r(obs, fcst_ens, remove_neg=False, remove_zero=False):
    """Calculate the pearson correlation coefficient between observed values and the ensemble mean.

    Parameters
    ----------
    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    Returns
    -------
    float
        The pearson correlation coefficient between the observed time series data and the ensemble
        mean.

    Examples
    --------
    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array

    >>> ensemble_array = (np.random.rand(100, 52) + 1) * 100
    >>> observed_array = (np.random.rand(100) + 1) * 100

    Computing the MSE between the ensemble mean and the observed data

    >>> em.ens_pearson_r(obs=observed_array, fcst_ens=ensemble_array)
    -0.13236871294739733

    """
    # Treating data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    fcst_ens_mean = np.mean(fcst_ens, axis=1)

    return pearson_r(fcst_ens_mean, obs)


def ens_crps(obs, fcst_ens, adj=np.nan, remove_neg=False, remove_zero=False, llvm=True):
    """Calculate the ensemble-adjusted Continuous Ranked Probability Score (CRPS)

    Parameters
    ----------
    obs: 1D ndarray
        array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    adj: float or int
        A positive number representing ensemble size for which the scores should be
        adjusted. If np.nan (default) scores will not be adjusted. This value can be ‘np.inf‘, in
        which case the adjusted (or fair) crps values will be calculated as per equation 6 in
        Leutbecher et al. (2018).

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    llvm: bool
        If true (default) the crps will be calculated using the numba module, which uses the LLVM compiler
        infrastructure for enhanced performance. If this is not wanted, then set it to false for a pure python
        implementation.

    Returns
    -------
    dict
        Dictionary contains two keys, crps and crpsMean. The value of crps is a list of the crps
        values. Note that if the ensemble forecast or the observed values contained NaN or inf
        values, or negative or zero values if specified, these start dates will not show up in the
        crps values. The crpsMean value is the arithmatic mean of the crps values.

    Examples
    --------

    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array with all random numbers

    >>> ens_array_random = (np.random.rand(15, 52) + 1) * 100  # 52 Ensembles
    >>> obs_array_random = (np.random.rand(15) + 1) * 100

    Creating an observed 1D array and an ensemble 2D array with noise.

    >>> noise = np.random.normal(scale=1, size=(15, 52))
    >>> x = np.linspace(1, 10, 15)
    >>> observed_array = np.sin(x) + 10
    >>> ensemble_array_noise = (np.ones((15, 52)).T * observed_array).T + noise  # 52 Ensembles

    Computing the crps values between the ensemble mean and the observed data with the
    random data. Note that the crps is relatively high because it is random.

    >>> crps_dictionary_rand = em.ens_crps(obs_array_random, ens_array_random)
    >>> print(crps_dictionary_rand['crps'])
    [ 7.73360237  9.59248626 34.46719655 30.10271075  7.451665   16.07882352
     14.59543529  8.55181637 15.4833089   8.32422363 16.55108154 19.20821296
      8.39452279 12.59949378 27.82543302]
    >>> crps_dictionary_rand['crpsMean']
    15.797334183277709

    Computing the crps values between the ensemble mean and the observed data with noise
    in the ensemble data. Note that the crps values are better because the forecast is closer to
    observed values.

    >>> crps_dictionary_noise = em.ens_crps(obs=observed_array, fcst_ens=ensemble_array_noise)
    >>> print(crps_dictionary_noise['crps'])
    [0.26921152 0.21388687 0.24927151 0.26047667 0.30234843 0.1996493
     0.2779844  0.29478927 0.275383   0.25682693 0.21485236 0.22824711
     0.2813889  0.21264652 0.18141063]
    >>> crps_dictionary_noise['crpsMean']
    0.24789156041214638

    References
    ----------
    - Gneiting, T. and Raftery, A. E. (2007) Strictly proper scoring rules,
      prediction and estimation, J. Amer. Stat. Asoc., 102, 359-378.
    - Leutbecher, M. (2018) Ensemble size: How suboptimal is less than infinity?,
      Q. J. R. Meteorol., Accepted.
    - Ferro CAT, Richardson SR, Weigel AP (2008) On the effect of ensemble size on the discrete and
      continuous ranked probability scores. Meteorological Applications. doi: 10.1002/met.45
    - Stefan Siegert (2017). SpecsVerification: Forecast Verification Routines for
      Ensemble Forecasts of Weather and Climate. R package version 0.5-2.
      https://CRAN.R-project.org/package=SpecsVerification
    """
    # Treating the Data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    rows = obs.size
    cols = fcst_ens.shape[1]

    col_len_array = np.ones(rows) * cols
    sad_ens_half = np.zeros(rows)
    sad_obs = np.zeros(rows)
    crps = np.zeros(rows)

    if llvm:
        crps = numba_crps(
            fcst_ens, obs, rows, cols, col_len_array, sad_ens_half, sad_obs, crps, np.float64(adj)
        )
    else:
        crps = python_crps(
            fcst_ens, obs, rows, cols, col_len_array, sad_ens_half, sad_obs, crps, np.float64(adj)
        )

    # Calc mean crps as simple mean across crps[i]
    crps_mean = np.mean(crps)

    # Output array as a dictionary
    output = {'crps': crps, 'crpsMean': crps_mean}

    return output


@jit(nopython=True, parallel=True)
def numba_crps(ens, obs, rows, cols, col_len_array, sad_ens_half, sad_obs, crps, adj):
    for i in prange(rows):
        the_obs = obs[i]
        the_ens = ens[i, :]
        the_ens = np.sort(the_ens)
        sum_xj = 0.
        sum_jxj = 0.

        j = 0
        while j < cols:
            sad_obs[i] += np.abs(the_ens[j] - the_obs)
            sum_xj += the_ens[j]
            sum_jxj += (j + 1) * the_ens[j]
            j += 1

        sad_ens_half[i] = 2.0 * sum_jxj - (col_len_array[i] + 1) * sum_xj

    if np.isnan(adj):
        for i in range(rows):
            crps[i] = sad_obs[i] / col_len_array[i] - sad_ens_half[i] / \
                      (col_len_array[i] * col_len_array[i])
    elif adj > 1:
        for i in range(rows):
            crps[i] = sad_obs[i] / col_len_array[i] - sad_ens_half[i] / \
                      (col_len_array[i] * (col_len_array[i] - 1)) * (1 - 1 / adj)
    elif adj == 1:
        for i in range(rows):
            crps[i] = sad_obs[i] / col_len_array[i]
    else:
        for i in range(rows):
            crps[i] = np.nan

    return crps


def python_crps(ens, obs, rows, cols, col_len_array, sad_ens_half, sad_obs, crps, adj):
    for i in prange(rows):
        the_obs = obs[i]
        the_ens = ens[i, :]
        the_ens = np.sort(the_ens)
        sum_xj = 0.
        sum_jxj = 0.

        j = 0
        while j < cols:
            sad_obs[i] += np.abs(the_ens[j] - the_obs)
            sum_xj += the_ens[j]
            sum_jxj += (j + 1) * the_ens[j]
            j += 1

        sad_ens_half[i] = 2.0 * sum_jxj - (col_len_array[i] + 1) * sum_xj

    if np.isnan(adj):
        for i in range(rows):
            crps[i] = sad_obs[i] / col_len_array[i] - sad_ens_half[i] / \
                      (col_len_array[i] * col_len_array[i])
    elif adj > 1:
        for i in range(rows):
            crps[i] = sad_obs[i] / col_len_array[i] - sad_ens_half[i] / \
                      (col_len_array[i] * (col_len_array[i] - 1)) * (1 - 1 / adj)
    elif adj == 1:
        for i in range(rows):
            crps[i] = sad_obs[i] / col_len_array[i]
    else:
        for i in range(rows):
            crps[i] = np.nan

    return crps


def crps_hersbach(obs, fcst_ens, remove_neg=False, remove_zero=False):
    """Calculate the the continuous ranked probability score (CRPS) as per equation 25-27 in
    Hersbach et al. (2000)

    Parameters
    ----------
    obs: 1D ndarry
        Array of observations for each start date
    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    Returns
    -------
    dict
        Dictionary contains a number of *experimental* outputs including:
            - ["crps"] 1D ndarray of crps values per n start dates.
            - ["crpsMean1"] arithmetic mean of crps values.
            - ["crpsMean2"] mean crps using eqn. 28 in Hersbach (2000).

    Notes
    -----
    **NaN and inf treatment:** If any value in obs or fcst_ens is NaN or inf, then the
    corresponding row in both fcst_ens (for all ensemble members) and in obs will be deleted.

    References
    ----------
    - Hersbach, H. (2000) Decomposition of the Continuous Ranked Porbability Score
      for Ensemble Prediction Systems, Weather and Forecasting, 15, 559-570.

    Examples
    --------

    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array with all random numbers

    >>> ens_array_random = (np.random.rand(15, 52) + 1) * 100
    >>> obs_array_random = (np.random.rand(15) + 1) * 100

    Creating an observed 1D array and an ensemble 2D array with noise.

    >>> noise = np.random.normal(scale=1, size=(15, 52))
    >>> x = np.linspace(1, 10, 15)
    >>> observed_array = np.sin(x) + 10
    >>> ensemble_array_noise = (np.ones((15, 52)).T * observed_array).T + noise

    Computing the Hersbach CRPS values between the ensemble mean and the observed data with the
    random data.

    >>> crps_dictionary_rand = em.crps_hersbach(obs_array_random, ens_array_random)
    >>> print(crps_dictionary_rand['crps'])
    [ 7.73360237  9.59248626 34.46719655 30.10271075  7.451665   16.07882352
     14.59543529  8.55181637 15.4833089   8.32422363 16.55108154 19.20821296
      8.39452279 12.59949378 27.82543302]
    >>> crps_dictionary_rand['crpsMean1']
    15.797334183277723
    >>> crps_dictionary_rand['crpsMean2']
    15.797334183277725

    Computing the Hersbach CRPS values between the ensemble mean and the observed data with noise
    in the ensemble data.

    >>> crps_dictionary_noise = em.crps_hersbach(obs=observed_array, fcst_ens=ensemble_array_noise)
    >>> print(crps_dictionary_noise['crps'])
    [0.26921152 0.21388687 0.24927151 0.26047667 0.30234843 0.1996493
     0.2779844  0.29478927 0.275383   0.25682693 0.21485236 0.22824711
     0.2813889  0.21264652 0.18141063]
    >>> crps_dictionary_noise['crpsMean1']
    0.24789156041214705
    >>> crps_dictionary_noise['crpsMean2']
    0.24789156041214705

    """

    # Treating the Data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    # Set parameters
    n = fcst_ens.shape[0]  # number of forecast start dates
    m = fcst_ens.shape[1]  # number of ensemble members

    # Create vector of pi's
    p = np.linspace(0, m, m + 1)
    pi = p / m

    crps = np.zeros(n)
    # crpsAdj = np.zeros(n)

    # Matrices for alpha and beta in CRPS decomposition
    a_mat = np.zeros(shape=(n, m + 1))
    b_mat = np.zeros(shape=(n, m + 1))

    # Loop fcst start times
    for i in range(n):

        # Initialise vectors for storing output
        a = np.zeros(m - 1)
        b = np.zeros(m - 1)

        # Verifying analysis (or obs)
        xa = obs[i]

        # Ensemble fcst CDF
        x = np.sort(fcst_ens[i, :])

        # Deal with 0 < i < m [So, will loop 50 times for m = 51]
        for j in range(m - 1):

            # Rule 1
            if xa > x[j + 1]:
                a[j] = x[j + 1] - x[j]
                b[j] = 0

            # Rule 2
            if x[j] < xa < x[j + 1]:
                a[j] = xa - x[j]
                b[j] = x[j + 1] - xa

            # Rule 3
            if xa < x[j]:
                a[j] = 0
                b[j] = x[j + 1] - x[j]

        # Deal with outliers for i = 0, and i = m,
        # else a & b are 0 for non-outliers
        if xa < x[0]:
            a1 = 0
            b1 = x[0] - xa
        else:
            a1 = 0
            b1 = 0

        # Upper outlier (rem m-1 is for last member m, but ptyhon is 0-based indexing)
        if xa > x[m - 1]:
            am = xa - x[m - 1]
            bm = 0
        else:
            am = 0
            bm = 0

        # Combine full a & b vectors including outlier
        a = np.insert(a, 0, a1)
        a = np.append(a, am)
        b = np.insert(b, 0, b1)
        b = np.append(b, bm)

        # Populate a_mat and b_mat
        a_mat[i, :] = a
        b_mat[i, :] = b

        # Calc crps for individual start times
        crps[i] = ((a * pi ** 2) + (b * (1 - pi) ** 2)).sum()

    # Calc mean crps as simple mean across crps[i]
    crps_mean_method1 = crps.mean()

    # Calc mean crps across all start times from eqn. 28 in Hersbach (2000)
    abar = np.mean(a_mat, 0)
    bbar = np.mean(b_mat, 0)
    crps_mean_method2 = ((abar * pi ** 2) + (bbar * (1 - pi) ** 2)).sum()

    # Output array as a dictionary
    output = {'crps': crps, 'crpsMean1': crps_mean_method1,
              'crpsMean2': crps_mean_method2}

    return output


def crps_kernel(obs, fcst_ens, remove_neg=False, remove_zero=False):
    """Compute the kernel representation of the continuous ranked probability score (CRPS).

    Calculates the kernel representation of the continuous ranked probability score (CRPS) as per
    equation 3 in Leutbecher et al. (2018) and the adjusted (or fair) crps as per equation 6 in the
    same paper. Note that it was Gneiting and Raftery (2007) who show the kernel representation as
    calculated here is equivalent to the standard definition based on the integral
    over the squared error of the cumulative distribution.

    Parameters
    ----------
    obs: 1D ndarray
        array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    remove_neg: bool
        If True, when a negative value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.

    remove_zero: bool
        If true, when a zero value is found at the i-th position in the observed OR ensemble
        array, the i-th value of the observed AND ensemble array are removed before the
        computation.


    Returns
    -------
    dict
        Dictionary of outputs includes:
            - ["crps"] 1D ndarray with crps values of length n.
            - ["crpsAdjusted"] 1D ndarray with adjusted crps values of length n.
            - ["crpsMean"] Arithmetic mean of crps values as a float.
            - ["crpsAdjustedMean"] Arithmetic mean of adjusted crps values as a float.

    Notes
    -----
    **NaN treatment:** If any start date in obs is NaN, then the corresponding row in fcst
    (for all ensemble members) will also be deleted.

    Examples
    --------

    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array with all random numbers

    >>> ens_array_random = (np.random.rand(15, 52) + 1) * 100
    >>> obs_array_random = (np.random.rand(15) + 1) * 100

    Creating an observed 1D array and an ensemble 2D array with noise.

    >>> noise = np.random.normal(scale=1, size=(15, 52))
    >>> x = np.linspace(1, 10, 15)
    >>> observed_array = np.sin(x) + 10
    >>> ensemble_array_noise = (np.ones((15, 52)).T * observed_array).T + noise

    Computing the Hersbach CRPS values between the ensemble mean and the observed data with the
    random data.

    >>> crps_dictionary_rand = em.crps_kernel(obs_array_random, ens_array_random)
    >>> print(crps_dictionary_rand['crps'])
    [ 7.73360237  9.59248626 34.46719655 30.10271075  7.451665   16.07882352
     14.59543529  8.55181637 15.4833089   8.32422363 16.55108154 19.20821296
      8.39452279 12.59949378 27.82543302]
    >>> print(crps_dictionary_rand['crpsAdjusted'])
    [ 7.43000827  9.29100065 34.14067524 29.76359191  7.14776152 15.75147589
     14.25192856  8.23647876 15.19419171  8.05998301 16.26113448 18.90686679
      8.09725139 12.24021268 27.45673444]
    >>> crps_dictionary_rand['crpsMean']
    15.797334183277723
    >>> crps_dictionary_rand['crpsAdjustedMean']
    15.481953018707593

    Computing the Hersbach CRPS values between the ensemble mean and the observed data with noise
    in the ensemble data.

    >>> crps_dictionary_noise = em.crps_kernel(obs=observed_array, fcst_ens=ensemble_array_noise)
    >>> print(crps_dictionary_noise['crps'])
    [0.26921152 0.21388687 0.24927151 0.26047667 0.30234843 0.1996493
     0.2779844  0.29478927 0.275383   0.25682693 0.21485236 0.22824711
     0.2813889  0.21264652 0.18141063]
    >>> print(crps_dictionary_noise['crpsAdjusted'])
    [0.25850726 0.20482797 0.23908004 0.25032814 0.28996894 0.18961905
     0.2670867  0.2821429  0.2634554  0.24573507 0.20457832 0.21730326
     0.26951946 0.2034818  0.17198615]
    >>> crps_dictionary_noise['crpsMean']
    0.2478915604121471
    >>> crps_dictionary_noise['crpsAdjustedMean']
    0.23717469718824744

    References
    ----------
    - Gneiting, T. and Raftery, A. E. (2007) Strictly proper scoring rules,
      prediction and estimation, J. Amer. Stat. Asoc., 102, 359-378.
    - Leutbecher, M. (2018) Ensemble size: How suboptimal is less than infinity?,
      Q. J. R. Meteorol., Accepted.
    """

    # Treatment of data
    obs, fcst_ens = treat_data(obs, fcst_ens, remove_neg=remove_neg, remove_zero=remove_zero)

    # Set parameters
    n = fcst_ens.shape[0]  # number of forecast start dates
    m = fcst_ens.shape[1]  # number of ensemble members

    # Initialise vectors for storing output
    t1 = np.zeros(n)
    t2 = np.zeros(n)
    crps = np.zeros(n)
    crps_adj = np.zeros(n)

    # Loop through start dates
    for i in range(n):

        t1[i] = abs(fcst_ens[i] - obs[i]).sum()

        # Initialise a vec for storing absolute errors for each ensemble pair
        vec = np.zeros(m)

        # Loop through ensemble members
        for j in range(m):
            vec[j] = abs(fcst_ens[i, j] - np.delete(fcst_ens[i, :], j)).sum()
            # vec[j] = abs(fcst[i, j] - fcst[i, :]).sum()

        t2[i] = vec.sum()

        # First term (t1) is the MAE of the ensemble members; Second term (t2)
        # is ensemble spread in terms of absolute difference between all pairs
        # of members
        crps[i] = (1. / m * t1[i]) - (1. / (2 * (m ** 2)) * t2[i])  # kernel representation of crps
        crps_adj[i] = (1. / m * t1[i]) - (
                1. / (2 * m * (m - 1)) * t2[i])  # kernel representation of adjusted crps

    # Calculate mean crps
    crps_mean = crps.mean()
    crps_adj_mean = crps_adj.mean()

    # Output two arrays as a dictionary
    output = {'crps': crps, 'crpsAdjusted': crps_adj, 'crpsMean': crps_mean,
              'crpsAdjustedMean': crps_adj_mean}

    return output


def ens_brier(fcst_ens=None, obs=None, threshold=None, fcst_ens_bin=None, obs_bin=None, adj=None):
    """Calculate the ensemble-adjusted Brier Score.

    Parameters
    ----------
    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    threshold: float
        The threshold for an event (e.g. if the event is a 100 year flood, the streamflow value
        that a 100 year flood would have to exceed.

    fcst_ens_bin: 1D ndarray
        Binary array of observations for each start date. 1 for an event and 0 for a non-event.

    obs_bin: 2D ndarray
        Binary array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members. 1 for an event and 0 for a non-event.

    adj: float or int
        A positive number representing ensemble size for which the scores should be
        adjusted. If None (default) scores will not be adjusted. This value can be ‘np.inf‘, in
        which case the adjusted (or fair) Brier scores will be calculated.

    Returns
    -------
    1D ndarray
        Array of length with the ensemble-adjusted Brier scores. Length may not equal n if data
        has been removed due to NaN or Inf values.

    Notes
    -----
    **NaN and inf treatment:** If any value in obs or fcst_ens is NaN or inf, then the
    corresponding row in both fcst_ens (for all ensemble members) and in obs will be deleted.

    Examples
    --------
    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)  # For reproducibility

    Creating an observed 1D array and an ensemble 2D array

    >>> ensemble_array = (np.random.rand(15, 52) + 1) * 100  # 52 Ensembles
    >>> observed_array = (np.random.rand(15) + 1) * 100

    Computing the ensemble-adjusted Brier score between the ensemble mean and the observed data.

    >>> print(em.ens_brier(obs=observed_array, fcst_ens=ensemble_array, threshold=175))
    [0.08321006 0.05325444 0.53402367 0.45303254 0.02995562 0.08321006
     0.08321006 0.03698225 0.02366864 0.0625     0.04474852 0.71597633
     0.04474852 0.04474852 0.09467456]
    >>> np.mean(em.ens_brier(obs=observed_array, fcst_ens=ensemble_array, threshold=175))
    0.15919625246548325

    When we manually create binary data we get the same result

    >>> ensemble_array_bin = (ensemble_array > 175).astype(np.int)
    >>> observed_array_bin = (observed_array > 175).astype(np.int)
    >>> print(em.ens_brier(obs_bin=observed_array_bin, fcst_ens_bin=ensemble_array_bin))
    [0.08321006 0.05325444 0.53402367 0.45303254 0.02995562 0.08321006
     0.08321006 0.03698225 0.02366864 0.0625     0.04474852 0.71597633
     0.04474852 0.04474852 0.09467456]
    >>> np.mean(em.ens_brier(obs_bin=observed_array_bin, fcst_ens_bin=ensemble_array_bin))
    0.15919625246548325

    References
    ----------
    - Ferro CAT, Richardson SR, Weigel AP (2008) On the effect of ensemble size on the discrete and
      continuous ranked probability scores. Meteorological Applications. doi: 10.1002/met.45
    - Stefan Siegert (2017). SpecsVerification: Forecast Verification Routines for
      Ensemble Forecasts of Weather and Climate. R package version 0.5-2.
      https://CRAN.R-project.org/package=SpecsVerification
    """

    if obs_bin is not None and fcst_ens_bin is not None:
        pass
    elif fcst_ens is not None and obs is not None and threshold is not None:
        # Convert the observed data and forecast data to binary data.
        obs_bin = (obs > threshold).astype(np.int)
        fcst_ens_bin = (fcst_ens > threshold).astype(np.int)
    else:
        raise RuntimeError(" You must either supply fcst_ens, obs, and threshold or you must "
                           "supply fcst_ens_bin and obs_bin.")

    # Treat missing data and warn users of columns being removed
    obs_bin, fcst_ens_bin = treat_data(obs_bin, fcst_ens_bin, remove_neg=False,
                                       remove_zero=False)

    # Count number of ensemble members that predict the event
    i = np.sum(fcst_ens_bin, axis=1)

    # Calculate ensemble size
    num_cols = fcst_ens_bin.shape[1]

    # No correction for ensemble size is performed if True
    if adj is None:
        adj = num_cols

    # calculate ensemble-adjusted brier scores
    br = (i / num_cols - obs_bin) ** 2 - i * (num_cols - i) / num_cols / \
         (num_cols - 1) * (1 / num_cols - 1 / adj)

    # return the vector of brier scores
    return br


def auroc(fcst_ens=None, obs=None, threshold=None, fcst_ens_bin=None, obs_bin=None):
    """Calculates Area Under the Relative Operating Characteristic curve (AUROC)
    for a forecast and its verifying binary observation, and estimates the variance of the AUROC

    Parameters
    ----------
    obs: 1D ndarray
        Array of observations for each start date.

    fcst_ens: 2D ndarray
        Array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members.

    threshold: float
        The threshold for an event (e.g. if the event is a 100 year flood, the streamflow value
        that a 100 year flood would have to exceed.

    fcst_ens_bin: 1D ndarray
        Binary array of observations for each start date. 1 for an event and 0 for a non-event.

    obs_bin: 2D ndarray
        Binary array of ensemble forecast of dimension n x M, where n = number of start dates and
        M = number of ensemble members. 1 for an event and 0 for a non-event.

    Notes
    -----
    **NaN and inf treatment:** If any value in obs or fcst_ens is NaN or inf, then the
    corresponding row in both fcst_ens (for all ensemble members) and in obs will be deleted. A
    warning will be shown that informs the user of the rows that have been removed.

    Returns
    -------
    1D ndarray
        An array of two elements, the AUROC and the estimated variance, respectively.

    Examples
    --------

    >>> import numpy as np
    >>> import hydrostats.ens_metrics as em
    >>> np.random.seed(3849590438)

    Creating an observed 1D array and an ensemble 2D array with all random numbers

    >>> ens_array_random = (np.random.rand(100, 52) + 1) * 100
    >>> obs_array_random = (np.random.rand(100) + 1) * 100

    Creating an observed 1D array and an ensemble 2D array with noise.

    >>> noise = np.random.normal(scale=1, size=(100, 52))
    >>> x = np.linspace(1, 10, 100)
    >>> observed_array = np.sin(x) + 10
    >>> ensemble_array_noise = (np.ones((100, 52)).T * observed_array).T + noise

    Calculating the ROC with random values. Note that the area under the curve is close to 0.5
    because the data is random.

    >>> print(em.auroc(obs=obs_array_random, fcst_ens=ens_array_random, threshold=175))
    [0.45560516 0.06406262]

    Calculating the ROC with noise in the forecast values. Note that the ROC value is high because
    the forecast is more accurate.

    >>> print(em.auroc(obs=observed_array, fcst_ens=ensemble_array_noise, threshold=10))
    [0.99137931 0.00566026]

    References
    ----------
    - DeLong et al (1988): Comparing the Areas under Two or More Correlated Receiver Operating
      Characteristic Curves: A Nonparametric Approach. Biometrics. doi: 10.2307/2531595
    - Sun and Xu (2014): Fast Implementation of DeLong's Algorithm for Comparing the Areas Under
      Correlated Receiver Operating Characteristic Curves. IEEE Sign Proc Let 21(11).
      doi: 10.1109/LSP.2014.2337313
    - Stefan Siegert (2017). SpecsVerification: Forecast Verification Routines for Ensemble
      Forecasts of Weather and Climate. R package version 0.5-2.
      https://CRAN.R-project.org/package=SpecsVerification

    """
    if obs_bin is not None and fcst_ens_bin is not None and fcst_ens is None and obs is None and threshold is None:
        pass
    elif fcst_ens is not None and obs is not None and threshold is not None and obs_bin is None and \
            fcst_ens_bin is None:
        # Convert the observed data and forecast data to binary data.
        obs_bin = (obs > threshold).astype(np.int)
        fcst_ens_bin = (fcst_ens > threshold).astype(np.int)
    else:
        raise RuntimeError(" You must either supply fcst_ens, obs, and threshold or you must "
                           "supply fcst_ens_bin and obs_bin.")

    obs_bin, fcst_ens_bin = treat_data(obs_bin, fcst_ens_bin, remove_neg=False, remove_zero=False)

    if np.all(fcst_ens_bin == 0) or np.all(fcst_ens_bin == 1) or np.all(obs_bin == 0) or np.all(obs_bin == 1):
        raise RuntimeError("Both arrays need at least one event and one non-event, otherwise, "
                           "division by zero will occur!")

    ens_forecast_means = np.mean(fcst_ens_bin, axis=1)

    results = auroc_numba(ens_forecast_means, obs_bin)

    return results


@jit(nopython=True)
def auroc_numba(fcst, obs):
    num_start_dates = obs.size

    i_ord = fcst.argsort()

    sum_v = 0.
    sum_v2 = 0.
    sum_w = 0.
    sum_w2 = 0.
    n = 0
    m = 0
    i = 0

    x = 1
    y = 0

    while True:
        nn = mm = 0
        while x > y:
            j = i_ord[i]
            if obs[j]:
                mm += 1
            else:
                nn += 1
            if i == num_start_dates - 1:
                break
            jp1 = i_ord[i + 1]
            if fcst[j] != fcst[jp1]:
                break
            i += 1
        sum_w += nn * (m + mm / 2.0)
        sum_w2 += nn * (m + mm / 2.0) * (m + mm / 2.0)
        sum_v += mm * (n + nn / 2.0)
        sum_v2 += mm * (n + nn / 2.0) * (n + nn / 2.0)
        n += nn
        m += mm
        i += 1
        if i >= num_start_dates:
            break

    theta = sum_v / (m * n)
    v = sum_v2 / ((m - 1) * n * n) - sum_v * sum_v / (m * (m - 1) * n * n)
    w = sum_w2 / ((n - 1) * m * m) - sum_w * sum_w / (n * (n - 1) * m * m)

    sd_auc = np.sqrt(v / m + w / n)

    return np.array([theta, sd_auc])


def skill_score(scores, bench_scores, perf_score, eff_sample_size=None, remove_nan_inf=False):
    """Calculate the skill score of the given function.

    Parameters
    ----------

    scores: float or ndarray
        The verification scores, or the mean of the verification scores in an ndarray (float).

    bench_scores: float or ndarray
        The reference or benchmark verification scores, or the mean of the benchmark scores (float).

    perf_score: int or float
        The perfect score of the score, typically 1 or 0.

    eff_sample_size: float
        The effective sample size of the data to be used when estimating the sampling uncertainty. Default is None,
        which will set the eff_sample_size to the length of scores.

    remove_nan_inf: bool
        If True, removes NaN and Inf values in the scores if they exist, pairwise. If False (default), the function
        will raise an exception.

    Returns
    -------
    dict
        Dictionary containing: {"skillScore": Float, the skill score, "standardDeviation": Float, the estimated standard
        deviation of the skill score} If the scores and bench scores given were floats, the standard deviation will be
        NaN.

    References
    ----------
    - Stefan Siegert (2017). SpecsVerification: Forecast Verification Routines for Ensemble Forecasts of Weather and
      Climate. R package version 0.5-2. https://CRAN.R-project.org/package=SpecsVerification

    Examples
    --------

    """
    # Check data
    assert np.isfinite(perf_score), 'The perfect score is not finite.'
    if eff_sample_size is not None:
        assert eff_sample_size > 0 and np.isfinite(eff_sample_size), 'The effective sample size must be finite and ' \
                                                                     'greater than 0.'

    if isinstance(scores, np.ndarray) and isinstance(bench_scores, np.ndarray):
        assert scores.size == bench_scores.size, 'The scores and benchmark scores are not the same length'

        # Making a copy to avoid altering original scores
        scores_copy = np.copy(scores)
        bench_scores_copy = np.copy(bench_scores)

        # Removing NaN and Inf if requested
        if remove_nan_inf:

            all_treatment_array = np.ones(scores_copy.size, dtype=bool)

            if np.any(np.isnan(scores_copy)) or np.any(np.isnan(bench_scores_copy)):
                nan_indices_fcst = ~np.isnan(scores_copy)
                nan_indices_obs = ~np.isnan(bench_scores_copy)
                all_nan_indices = np.logical_and(nan_indices_fcst, nan_indices_obs)
                all_treatment_array = np.logical_and(all_treatment_array, all_nan_indices)

                warnings.warn("Row(s) {} contained NaN values and the row(s) have been "
                              "removed for the calculation (Rows are zero indexed).".format(np.where(~all_nan_indices)[0]),
                              UserWarning)

            if np.any(np.isinf(scores_copy)) or np.any(np.isinf(bench_scores_copy)):
                inf_indices_fcst = ~(np.isinf(scores_copy))
                inf_indices_obs = ~np.isinf(bench_scores_copy)
                all_inf_indices = np.logical_and(inf_indices_fcst, inf_indices_obs)
                all_treatment_array = np.logical_and(all_treatment_array, all_inf_indices)

                warnings.warn(
                    "Row(s) {} contained Inf or -Inf values and the row(s) have been removed for the calculation (Rows "
                    "are zero indexed).".format(np.where(~all_inf_indices)[0]),
                    UserWarning
                )

            scores_copy = scores_copy[all_treatment_array]
            bench_scores_copy = bench_scores_copy[all_treatment_array]

        else:  # If User didn't want to remove NaN and Inf
            if np.any(np.isnan(scores_copy)) or np.any(np.isnan(bench_scores_copy)):
                nan_indices_fcst = ~np.isnan(scores_copy)
                nan_indices_obs = ~np.isnan(bench_scores_copy)
                all_nan_indices = np.logical_and(nan_indices_fcst, nan_indices_obs)

                raise RuntimeError("Row(s) {} contained NaN values "
                                   "(Rows are zero indexed).".format(np.where(~all_nan_indices)[0]))

            if np.any(np.isinf(scores_copy)) or np.any(np.isinf(bench_scores_copy)):
                inf_indices_fcst = ~(np.isinf(scores_copy))
                inf_indices_obs = ~np.isinf(bench_scores_copy)
                all_inf_indices = np.logical_and(inf_indices_fcst, inf_indices_obs)

                raise RuntimeError("Row(s) {} contained Inf or -Inf values "
                                   "(Rows are zero indexed).".format(np.where(~all_inf_indices)[0]))

        # Handle effective sample size
        if eff_sample_size is None:
            eff_sample_size = scores.size

        # calculate mean scores, shift by score.perf
        score = np.mean(scores_copy) - perf_score
        bench_score = np.mean(bench_scores_copy) - perf_score

        # calculate skill score
        skillscore = 1 - score / bench_score

        # calculate auxiliary quantities
        var_score = np.var(scores_copy, ddof=1)
        var_bench_score = np.var(bench_scores_copy, ddof=1)
        cov_score = np.cov(scores_copy, bench_scores_copy)[0, 1]

        # Calculate skill score standard deviation by error propagation
        def sqrt_na(z):
            if z < 0:
                z = np.nan

            return np.sqrt(z)

        sqrt_na_val = sqrt_na(
            var_score / bench_score ** 2 + var_bench_score * score ** 2 / bench_score ** 4 - 2 * cov_score *
            score / bench_score ** 3
        )

        skillscore_sigma = (1 / np.sqrt(eff_sample_size)) * sqrt_na_val

        # Set skillscore_sigma to NaN if not finite
        if not np.isfinite(skillscore_sigma):
            skillscore_sigma = np.nan

    elif isinstance(scores, float) and isinstance(bench_scores, float):

        # shift mean scores by perfect score
        score = scores - perf_score
        bench_score = bench_scores - perf_score

        # calculate skill score
        skillscore = 1 - score / bench_score

        skillscore_sigma = np.nan

    else:
        raise RuntimeError("The scores and benchmark_scores must either both be ndarrays or both be floats.")

    return_dict = {
        'skillScore': skillscore,
        'standardDeviation': skillscore_sigma
    }

    return return_dict


def treat_data(obs, fcst_ens, remove_zero, remove_neg):
    assert obs.ndim == 1, "obs is not a 1D numpy array."
    assert fcst_ens.ndim == 2, "fcst_ens is not a 2D numpy array."
    assert obs.size == fcst_ens[:, 0].size, "obs and fcst_ens do not have the same amount " \
                                            "of start dates."

    # Give user warning, but let run, if eith obs or fcst are all zeros
    if obs.sum() == 0 or fcst_ens.sum() == 0:
        warnings.warn("All zero values in either 'obs' or 'fcst', "
                      "function might run, but check if data OK.")

    all_treatment_array = np.ones(obs.size, dtype=bool)

    # Treat missing data in obs and fcst_ens, rows in fcst_ens or obs that contain nan values
    if np.any(np.isnan(obs)) or np.any(np.isnan(fcst_ens)):
        nan_indices_fcst = ~(np.any(np.isnan(fcst_ens), axis=1))
        nan_indices_obs = ~np.isnan(obs)
        all_nan_indices = np.logical_and(nan_indices_fcst, nan_indices_obs)
        all_treatment_array = np.logical_and(all_treatment_array, all_nan_indices)

        warnings.warn("Row(s) {} contained NaN values and the row(s) have been "
                      "removed (zero indexed).".format(np.where(~all_nan_indices)[0]))

    if np.any(np.isinf(obs)) or np.any(np.isinf(fcst_ens)):
        inf_indices_fcst = ~(np.any(np.isinf(fcst_ens), axis=1))
        inf_indices_obs = ~np.isinf(obs)
        all_inf_indices = np.logical_and(inf_indices_fcst, inf_indices_obs)
        all_treatment_array = np.logical_and(all_treatment_array, all_inf_indices)

        warnings.warn("Row(s) {} contained Inf or -Inf values and the row(s) have been "
                      "removed (zero indexed).".format(np.where(~all_inf_indices)[0]))

    # Treat zero data in obs and fcst_ens, rows in fcst_ens or obs that contain zero values
    if remove_zero:
        if (obs == 0).any() or (fcst_ens == 0).any():
            zero_indices_fcst = ~(np.any(fcst_ens == 0, axis=1))
            zero_indices_obs = ~(obs == 0)
            all_zero_indices = np.logical_and(zero_indices_fcst, zero_indices_obs)
            all_treatment_array = np.logical_and(all_treatment_array, all_zero_indices)

            warnings.warn("Row(s) {} contained zero values and the row(s) have been "
                          "removed (zero indexed).".format(np.where(~all_zero_indices)[0]))

    # Treat negative data in obs and fcst_ens, rows in fcst_ens or obs that contain negative values
    # warnings.filterwarnings("ignore")  # Ignore Runtime warnings for comparison
    if remove_neg:
        with np.errstate(invalid='ignore'):
            obs_bool = obs < 0
            fcst_ens_bool = fcst_ens < 0
        if obs_bool.any() or fcst_ens_bool.any():
            neg_indices_fcst = ~(np.any(fcst_ens_bool, axis=1))
            neg_indices_obs = ~obs_bool
            all_neg_indices = np.logical_and(neg_indices_fcst, neg_indices_obs)
            all_treatment_array = np.logical_and(all_treatment_array, all_neg_indices)

            # warnings.filterwarnings("always")  # Turn warnings back on

            warnings.warn("Row(s) {} contained negative values and the row(s) have been "
                          "removed (zero indexed).".format(np.where(~all_neg_indices)[0]))
        else:
            pass  # warnings.filterwarnings("always")  # Turn warnings back on
    else:
        pass  # warnings.filterwarnings("always")  # Turn warnings back on

    obs = obs[all_treatment_array]
    fcst_ens = fcst_ens[all_treatment_array, :]

    return obs, fcst_ens


if __name__ == "__main__":
    import time
    ensemble_array = np.load(r"tests/Files_for_tests/ensemble_array.npz")["arr_0.npy"]
    observed_array = np.load(r"tests/Files_for_tests/observed_array.npz")["arr_0.npy"]

    ens_bin = np.load(r"tests/Files_for_tests/ens_bin.npz")["arr_0.npy"]
    obs_bin = np.load(r"tests/Files_for_tests/obs_bin.npz")["arr_0.npy"]

    ensemble_array_bad_data = np.load(r"tests/Files_for_tests/ensemble_array_bad_data.npz")["arr_0.npy"]
    observed_array_bad_data = np.load(r"tests/Files_for_tests/observed_array_bad_data.npz")["arr_0.npy"]

    start = time.time()
    crps_numba = ens_crps(obs=observed_array, fcst_ens=ensemble_array)["crpsMean"]
    print(time.time() - start)

    start = time.time()
    crps_python = ens_crps(obs=observed_array, fcst_ens=ensemble_array, llvm=False)["crpsMean"]
    print(time.time() - start)

    assert np.isclose(crps_numba, crps_python)

    print(crps_numba, crps_python)



# python 3.6
# -*- coding: utf-8 -*-
"""
The metrics module contains all of the metrics included in the HydroErr package. These metrics measure hydrologic skill.
Each metric is contained in function, and every metric has the ability to treat missing values as
well as remove zero and negative values from the timeseries data.
"""
from HydroErr import *
from HydroErr.HydroErr import metric_names, metric_abbr, function_list


def list_of_metrics(metrics, sim_array, obs_array, abbr=False, mase_m=1, dmod_j=1,
                    nse_mod_j=1, h6_mhe_k=1, h6_ahe_k=1, h6_rmshe_k=1, d1_p_obs_bar_p=None,
                    lm_x_obs_bar_p=None, kge2009_s=(1, 1, 1), kge2012_s=(1, 1, 1), replace_nan=None, replace_inf=None, remove_neg=False,
                    remove_zero=False):
    if sim_array.ndim != 1 or obs_array.ndim != 1:
        raise RuntimeError("One or both of the ndarrays are not 1 dimensional.")
    if sim_array.size != obs_array.size:
        raise RuntimeError("The two ndarrays are not the same size.")

    metrics_list = []

    if not abbr:
        for metric in metrics:
            if metric == 'Mean Absolute Scaled Error':
                metrics_list.append(mase(sim_array, obs_array, m=mase_m,
                                         replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == 'Modified Index of Agreement':
                metrics_list.append(dmod(sim_array, obs_array, j=dmod_j,
                                         replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == 'Modified Nash-Sutcliffe Efficiency':
                metrics_list.append(nse_mod(sim_array, obs_array, j=nse_mod_j,
                                            replace_nan=replace_nan, replace_inf=replace_inf,
                                            remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == 'Legate-McCabe Efficiency Index':
                metrics_list.append(lm_index(sim_array, obs_array, obs_bar_p=lm_x_obs_bar_p,
                                             replace_nan=replace_nan, replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == 'Mean H6 Error':
                metrics_list.append(h6_mhe(sim_array, obs_array, k=h6_mhe_k,
                                           replace_nan=replace_nan, replace_inf=replace_inf,
                                           remove_neg=remove_neg, remove_zero=remove_zero
                                           ))

            elif metric == 'Mean Absolute H6 Error':
                metrics_list.append(h6_mahe(sim_array, obs_array, k=h6_ahe_k,
                                            replace_nan=replace_nan, replace_inf=replace_inf,
                                            remove_neg=remove_neg, remove_zero=remove_zero
                                            ))

            elif metric == 'Root Mean Square H6 Error':
                metrics_list.append(h6_rmshe(sim_array, obs_array, k=h6_rmshe_k,
                                             replace_nan=replace_nan, replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero
                                             ))

            elif metric == 'Legate-McCabe Index of Agreement':
                metrics_list.append(d1_p(sim_array, obs_array, obs_bar_p=d1_p_obs_bar_p,
                                         replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_neg=remove_neg, remove_zero=remove_zero
                                         ))
            elif metric == 'Kling-Gupta Efficiency (2009)':
                metrics_list.append(kge_2009(sim_array, obs_array, s=kge2009_s,
                                             replace_nan=replace_nan, replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero
                                             ))
            elif metric == 'Kling-Gupta Efficiency (2012)':
                metrics_list.append(kge_2012(sim_array, obs_array, s=kge2012_s,
                                             replace_nan=replace_nan, replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero
                                             ))
            else:
                index = metric_names.index(metric)
                metric_func = function_list[index]
                metrics_list.append(metric_func(sim_array, obs_array, replace_nan=replace_nan,
                                                replace_inf=replace_inf, remove_neg=remove_neg,
                                                remove_zero=remove_zero))

    else:
        for metric in metrics:
            if metric == 'MASE':
                metrics_list.append(mase(sim_array, obs_array, m=mase_m,
                                         replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == 'd (Mod.)':
                metrics_list.append(dmod(sim_array, obs_array, j=dmod_j,
                                         replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == 'NSE (Mod.)':
                metrics_list.append(nse_mod(sim_array, obs_array, j=nse_mod_j,
                                            replace_nan=replace_nan,
                                            replace_inf=replace_inf,
                                            remove_neg=remove_neg, remove_zero=remove_zero))

            elif metric == "E1'":
                metrics_list.append(lm_index(sim_array, obs_array, obs_bar_p=lm_x_obs_bar_p,
                                             replace_nan=replace_nan,
                                             replace_inf=replace_inf,
                                             remove_neg=remove_neg,
                                             remove_zero=remove_zero))

            elif metric == 'H6 (MHE)':
                metrics_list.append(h6_mhe(sim_array, obs_array, k=h6_mhe_k,
                                           replace_nan=replace_nan, replace_inf=replace_inf,
                                           remove_neg=remove_neg, remove_zero=remove_zero
                                           ))

            elif metric == 'H6 (AHE)':
                metrics_list.append(h6_mahe(sim_array, obs_array, k=h6_ahe_k,
                                            replace_nan=replace_nan, replace_inf=replace_inf,
                                            remove_neg=remove_neg, remove_zero=remove_zero
                                            ))

            elif metric == 'H6 (RMSHE)':
                metrics_list.append(h6_rmshe(sim_array, obs_array, k=h6_rmshe_k,
                                             replace_nan=replace_nan,
                                             replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero
                                             ))

            elif metric == "D1'":
                metrics_list.append(d1_p(sim_array, obs_array, obs_bar_p=d1_p_obs_bar_p,
                                         replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_neg=remove_neg, remove_zero=remove_zero
                                         ))
            elif metric == 'KGE (2009)':
                metrics_list.append(kge_2009(sim_array, obs_array, s=kge2009_s,
                                             replace_nan=replace_nan, replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero
                                             ))
            elif metric == 'KGE (2012)':
                metrics_list.append(kge_2012(sim_array, obs_array, s=kge2012_s,
                                             replace_nan=replace_nan, replace_inf=replace_inf,
                                             remove_neg=remove_neg, remove_zero=remove_zero
                                             ))
            else:
                index = metric_abbr.index(metric)
                metric_func = function_list[index]
                metrics_list.append(
                    metric_func(sim_array, obs_array, replace_nan=replace_nan,
                                replace_inf=replace_inf, remove_neg=remove_neg,
                                remove_zero=remove_zero))
    return metrics_list


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

The data module contains tools for preprocessing data. It allows users to merge timeseries, compute
daily and monthly summary statistics, and get seasonal periods of a time series.

"""
from __future__ import division
import pandas as pd
from numpy import inf, nan

__all__ = ['julian_to_gregorian', 'merge_data', 'daily_average', 'daily_std_error', 'daily_std_dev', 'monthly_average',
           'monthly_std_error', 'monthly_std_dev', 'remove_nan_df', 'seasonal_period']


def julian_to_gregorian(dataframe, frequency=None, inplace=False):
    """
    Converts the index of the merged dataframe from julian float values to gregorian datetime
    values.

    Parameters
    ----------
    dataframe: Pandas DataFrame
        A DataFrame with an index of type float

    frequency: string
        Optional. Sometimes when converting from julian to gregorian there will be rounding errors
        due to the inability of computers to store floats as perfect decimals. Providing the
        frequency will automatically attempt to round the dates. A list of all the frequencies pandas provides is found
        `here <http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases/>`_. Common frequencies
        include daily ("D") and hourly ("H").

    inplace: bool
        Default False. If True, will modify the index of the dataframe in place rather than
        creating a copy and returning the copy. Use when the time series are very long and making
        a copy would take a large amount of memory

    Returns
    -------
    Pandas DataFrame
        A pandas DataFrame with gregorian index.

    Examples
    --------

    >>> import pandas as pd
    >>> import hydrostats.data as hd
    >>> import numpy as np

    >>> # The julian dates in an array
    >>> julian_dates = np.array([2444239.5, 2444239.5416666665, 2444239.5833333335, 2444239.625,
    >>>                          2444239.6666666665, 2444239.7083333335, 2444239.75,
    >>>                          2444239.7916666665, 2444239.8333333335, 2444239.875])
    >>> # Creating a test dataframe
    >>> test_df = pd.DataFrame(data=np.random.rand(10, 2),  # Random data in the columns
    >>>                        columns=("Simulated Data", "Observed Data"),
    >>>                        index=julian_dates)
    >>> test_df
                  Simulated Data  Observed Data
    2.444240e+06        0.764719       0.126610
    2.444240e+06        0.372736       0.141392
    2.444240e+06        0.008645       0.686477
    2.444240e+06        0.656825       0.480444
    2.444240e+06        0.555247       0.869409
    2.444240e+06        0.643896       0.549590
    2.444240e+06        0.242720       0.799617
    2.444240e+06        0.432421       0.185760
    2.444240e+06        0.694631       0.136986
    2.444240e+06        0.700422       0.390415

    >>> # Making a new df with gregorian index
    >>> test_df_gregorian = hd.julian_to_gregorian(test_df)
    >>> test_df_gregorian
                              Simulated Data   Observed Data
    1980-01-01 00:00:00.000000	    0.585454	    0.457238
    1980-01-01 01:00:00.028800	    0.524764	    0.083464
    1980-01-01 01:59:59.971200	    0.516821	    0.416683
    1980-01-01 03:00:00.000000	    0.948483	    0.553874
    1980-01-01 04:00:00.028800	    0.492280	    0.232901
    1980-01-01 04:59:59.971200	    0.527967	    0.296395
    1980-01-01 06:00:00.000000	    0.650018	    0.212802
    1980-01-01 07:00:00.028800	    0.585592	    0.802971
    1980-01-01 07:59:59.971200	    0.448243	    0.665814
    1980-01-01 09:00:00.000000	    0.137395	    0.201721

    >>> # Rounding can be applied due to floating point inaccuracy
    >>> test_df_gregorian_rounded = julian_to_gregorian(test_df, frequency="H")  # Hourly Rounding Frequency
    >>> test_df_gregorian_rounded
                         Simulated Data  Observed Data
    1980-01-01 00:00:00        0.309527       0.938991
    1980-01-01 01:00:00        0.872284       0.497708
    1980-01-01 02:00:00        0.168046       0.225845
    1980-01-01 03:00:00        0.954494       0.275607
    1980-01-01 04:00:00        0.875885       0.194380
    1980-01-01 05:00:00        0.236849       0.992770
    1980-01-01 06:00:00        0.639346       0.029808
    1980-01-01 07:00:00        0.855828       0.903927
    1980-01-01 08:00:00        0.638805       0.916124
    1980-01-01 09:00:00        0.273430       0.443980

    >>> # The DataFrame can also be modified in place, increasing efficiency with large time series
    >>> julian_to_gregorian(test_df, inplace=True, frequency="H")
    >>> test_df
                         Simulated Data  Observed Data
    1980-01-01 00:00:00        0.309527       0.938991
    1980-01-01 01:00:00        0.872284       0.497708
    1980-01-01 02:00:00        0.168046       0.225845
    1980-01-01 03:00:00        0.954494       0.275607
    1980-01-01 04:00:00        0.875885       0.194380
    1980-01-01 05:00:00        0.236849       0.992770
    1980-01-01 06:00:00        0.639346       0.029808
    1980-01-01 07:00:00        0.855828       0.903927
    1980-01-01 08:00:00        0.638805       0.916124
    1980-01-01 09:00:00        0.273430       0.443980

    """

    if inplace:
        dataframe.index = pd.to_datetime(dataframe.index, origin="julian", unit="D")

        if frequency is not None:
            dataframe.index = dataframe.index.round(frequency)

    else:
        # Copying to avoid modifying the original dataframe
        return_df = dataframe.copy()

        # Converting the dataframe index from julian to gregorian
        return_df.index = pd.to_datetime(return_df.index, origin="julian", unit="D")

        if frequency is not None:
            return_df.index = return_df.index.round(frequency)

        return return_df


def merge_data(sim_fpath=None, obs_fpath=None, sim_df=None, obs_df=None, interpolate=None,
               column_names=('Simulated', 'Observed'), simulated_tz=None, observed_tz=None, interp_type='pchip',
               return_tz="Etc/UTC", julian=False, julian_freq=None):
    """Merges two dataframes or csv files, depending on the input.

    Parameters
    ----------
    sim_fpath: str
        The filepath to the simulated csv of data. Can be a url if the page is formatted correctly.
        The csv must be formatted with the dates in the left column and the data in the right
        column.

    obs_fpath: str
        The filepath to the observed csv. Can be a url if the page is formatted correctly.
        The csv must be formatted with the dates in the left column and the data in the right
        column.

    sim_df: DataFrame
        A pandas DataFrame containing the simulated data. Must be formatted with a datetime index
        and the simulated data values in column 0.

    obs_df: DataFrame
        A pandas DataFrame containing the simulated data. Must be formatted with a datetime index
        and the simulated data values in column 0.

    interpolate: str
        Must be either 'observed' or 'simulated'. Specifies which data set you would like to
        interpolate if interpolation is needed to properly merge the data.

    column_names: tuple of str
        Tuple of length two containing the column names that the user would like to set for the
        DataFrame that is returned. Note that the simulated data will be in the left column and the
        observed data will be in the right column

    simulated_tz: str
        The timezone of the simulated data. A full list of timezones can be found in the
        :ref:`timezones`.

    observed_tz: str
        The timezone of the simulated data. A full list of timezones can be found in the
        :ref:`timezones`.

    interp_type: str
        Which interpolation method to use. Uses the default pandas interpolater.
        Available types are found at
        http://pandas.pydata.org/pandas-docs/version/0.16.2/generated/pandas.DataFrame.interpolate.html

    return_tz: str
        What timezone the merged dataframe's index should be returned as. Default is 'Etc/UTC', which is recommended
        for simplicity.

    julian: bool
        If True, will parse the first column of the file to a datetime index from julian floating point time
        representation, this is only valid when supplying the sim_fpath and obs_fpath parameters. Users supplying two
        DataFrame objects must convert the index from Julian to Gregorian using the julian_to_gregorian function in this
        module

    julian_freq: str
        A string representing the frequency of the julian dates so that they can be rounded. See examples for usage.

    Notes
    -----
    The only acceptable time frequencies in the data are 15min, 30min, 45min, and any number of hours or
    days in between.

    There are three scenarios to consider when merging your data:

    1. The first scenario is that the timezones and the spacing of the time series matches
       (eg. 1 Day). In this case, you will want to leave the simulated_tz, observed_tz, and
       interpolate arguments empty, and the function will simply join the two csv's into a dataframe.
    2. The second scenario is that you have two time series with matching time zones but not
       matching spacing. In this case you will want to leave the simulated_tz and observed_tz empty,
       and use the interpolate argument to tell the function which time series you would like to
       interpolate to match the other time series.
    3. The third scenario is that you have two time series with different time zones and possibly
       different spacings. In this case you will want to fill in the simulated_tz, observed_tz, and
       interpolate arguments. This will then take timezones into account when interpolating
       the selected time series.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    """
    # Reading the data into dataframes if from file
    if sim_fpath is not None and obs_fpath is not None:
        # Importing data into a data-frame
        sim_df_copy = pd.read_csv(sim_fpath, delimiter=",", header=None, names=[column_names[0]],
                                  index_col=0, infer_datetime_format=True, skiprows=1)
        obs_df_copy = pd.read_csv(obs_fpath, delimiter=",", header=None, names=[column_names[1]],
                                  index_col=0, infer_datetime_format=True, skiprows=1)

        # Converting the index to datetime type
        if julian:
            julian_to_gregorian(sim_df_copy, frequency=julian_freq, inplace=True)
            julian_to_gregorian(obs_df_copy, frequency=julian_freq, inplace=True)
        else:
            sim_df_copy.index = pd.to_datetime(sim_df_copy.index, infer_datetime_format=True, errors='coerce')
            obs_df_copy.index = pd.to_datetime(obs_df_copy.index, infer_datetime_format=True, errors='coerce')

    elif sim_df is not None and obs_df is not None:
        # Checking to make sure that both dataframes have datetime indices if they are not read from file.
        if not isinstance(sim_df.index, pd.DatetimeIndex) and not isinstance(obs_df.index, pd.DatetimeIndex):
            raise RuntimeError("Both the obs_df and the sim_df need to have a datetime index.")

        # Copying the user supplied DataFrame objects
        sim_df_copy = sim_df.copy()
        obs_df_copy = obs_df.copy()

    else:
        raise RuntimeError('either sim_fpath and obs_fpath or sim_df and obs_df are required inputs.')

    # Checking to see if the necessary arguments in the function are fulfilled
    if simulated_tz is None and observed_tz is not None:
        raise RuntimeError('Either Both Timezones are required or neither')

    elif simulated_tz is not None and observed_tz is None:
        raise RuntimeError('Either Both Timezones are required or neither')

    elif simulated_tz is not None and observed_tz is not None and interpolate is None:
        raise RuntimeError("You must specify with the interpolate parameter whether to interpolate the 'simulated' "
                           "or 'observed' data.")

    elif simulated_tz is None and observed_tz is None and interpolate is None:
        # Scenario 1

        # Merging and joining the two DataFrames
        merged_df = pd.DataFrame.join(sim_df_copy, obs_df_copy).dropna()
        merged_df.columns = column_names

        return merged_df

    elif simulated_tz is None and observed_tz is None and interpolate is not None:
        # Scenario 2

        if interpolate == 'simulated':
            # Resampling and interpolating the observed data to match
            sim_df_copy = sim_df_copy.resample("15min").interpolate(interp_type)

        elif interpolate == 'observed':
            # Resampling and interpolating the observed data to match
            obs_df_copy = obs_df_copy.resample("15min").interpolate(interp_type)

        else:
            raise RuntimeError("The interpolate argument must be either 'simulated' or 'observed'.")

        # Merging and joining the two DataFrames
        merged_df = pd.DataFrame.join(sim_df_copy, obs_df_copy).dropna()
        merged_df.columns = column_names

        return merged_df

    elif simulated_tz is not None and observed_tz is not None and interpolate is not None:
        # Scenario 3

        # Convert the DateTime Index of both DataFrames to User Specified Timezones
        sim_df_copy.index = sim_df_copy.index.tz_localize(simulated_tz).tz_convert(return_tz)
        obs_df_copy.index = obs_df_copy.index.tz_localize(observed_tz).tz_convert(return_tz)

        if interpolate == 'simulated':
            # Resampling the simulated DataFrame to 15 minute time increments, then interpolating
            sim_df_copy = sim_df_copy.resample("15min").interpolate(interp_type)

        elif interpolate == 'observed':
            # Resampling the observed DataFrame to 15 minute time increments, then interpolating
            obs_df_copy = obs_df_copy.resample("15min").interpolate(interp_type)

        else:
            raise RuntimeError("You must specify the interpolation argument to be either 'simulated' or "
                               "'observed'.")

        # Merging and joining the two DataFrames
        merged_df = pd.DataFrame.join(sim_df_copy, obs_df_copy).dropna()
        merged_df.columns = column_names

        return merged_df


def daily_average(merged_data):
    """Calculates daily seasonal averages of the timeseries data in a DataFrame

    Parameters
    ----------

    merged_data: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        A pandas dataframe with a string type index of date representations and the daily seasonal averages as
        float values in the columns.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    >>> hd.daily_average(merged_df)
           Streamflow Prediction Tool       GLOFAS
    01/01                 7331.305278  7676.792460
    01/02                 7108.640746  7753.671798
    01/03                 6927.147740  7631.900453
    01/04                 6738.162886  7483.029897
    01/05                 6552.914171  7316.004227
    01/06                 6388.213829  7154.650963
    01/07                 6258.418600  7012.279722
                               ...          ...
    12/25                 8321.367143  8948.101821
    12/26                 8149.313143  8903.544978
    12/27                 7994.357429  8807.690639
    12/28                 7872.819143  8642.877365
    12/29                 7791.741143  8435.175677
    12/30                 7729.451143  8225.315074
    12/31                 7656.042286  8041.918136
    [366 rows x 2 columns]
    """
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m/%d"))
    return a.mean()


def daily_std_error(merged_data):
    """Calculates daily seasonal standard error of the timeseries data in a DataFrame

    Parameters
    ----------

    merged_data: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        A pandas dataframe with a string type index of date representations and the daily seasonal standard error as
        float values in the columns.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    >>> hd.daily_std_error(merged_df)
           Streamflow Prediction Tool      GLOFAS
    01/01                  558.189895  494.958042
    01/02                  553.290181  442.497656
    01/03                  535.002487  432.096928
    01/04                  514.511095  422.915060
    01/05                  489.287216  411.861086
    01/06                  463.321927  401.023620
    01/07                  441.666108  395.703128
                               ...         ...
    12/25                  613.876851  566.669886
    12/26                  589.424434  567.179646
    12/27                  582.957832  557.932109
    12/28                  581.465297  540.021918
    12/29                  573.949000  517.494155
    12/30                  560.993945  495.040565
    12/31                  546.904139  474.742075
    [366 rows x 2 columns]

    """
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m/%d"))
    return a.sem()


def daily_std_dev(merged_data):
    """Calculates daily seasonal standard deviation of the timeseries data in a DataFrame

    Parameters
    ----------

    merged_data: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        A pandas dataframe with a string type index of date representations and the daily seasonal standard deviation as
        float values in the columns.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    >>> hd.daily_std_dev(merged_df)
           Streamflow Prediction Tool       GLOFAS
    01/01                 3349.139373  2969.748253
    01/02                 3273.308852  2617.851437
    01/03                 3165.117397  2556.319898
    01/04                 3043.888685  2501.999235
    01/05                 2894.662206  2436.603046
    01/06                 2741.049485  2372.487729
    01/07                 2612.931931  2341.011275
                               ...          ...
    12/25                 3631.744428  3352.464257
    12/26                 3487.081980  3355.480036
    12/27                 3448.825041  3300.770870
    12/28                 3439.995086  3194.812751
    12/29                 3395.528078  3061.536706
    12/30                 3318.884936  2928.699478
    12/31                 3235.528520  2808.611992
    [366 rows x 2 columns]

    """
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m/%d"))
    return a.std()


def monthly_average(merged_data):
    """Calculates monthly seasonal averages of the timeseries data in a DataFrame

    Parameters
    ----------

    merged_data: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        A pandas dataframe with a string type index of date representations and the monthly seasonal averages as
        float values in the columns.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    >>> hd.monthly_average(merged_df)
            Streamflow Prediction Tool        GLOFAS
    01                 5450.558878   6085.033102
    02                 4178.249788   4354.072332
    03                 4874.788452   4716.785701
    04                 7682.920219   7254.073875
    05                13062.175899  11748.583189
    06                12114.431105  11397.032335
    07                 9461.472599   9598.017209
    08                 8802.643954   8708.876388
    09                10358.254219   9944.071882
    10                12968.474415  12671.180449
    11                13398.111010  13355.019167
    12                 9853.288608  10275.652887

    """
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m"))
    return a.mean()


def monthly_std_error(merged_data):
    """Calculates monthly seasonal standard error of the timeseries data in a DataFrame

    Parameters
    ----------

    merged_data: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        A pandas dataframe with a string type index of date representations and the monthly seasonal standard error as
        float values in the columns.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    >>> hd.monthly_std_error(merged_df)
        Streamflow Prediction Tool      GLOFAS
    01                   75.348943   71.206858
    02                   65.182159   58.347438
    03                   83.865980   72.919782
    04                  131.199766  123.486202
    05                  165.707528  139.586564
    06                  149.938998  136.119774
    07                  136.449337  115.436215
    08                  129.371343  110.101251
    09                  143.367894  123.798265
    10                  133.911782  114.008446
    11                  146.896826  116.443188
    12                  135.092750  117.958715

    """
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m"))
    return a.sem()


def monthly_std_dev(merged_data):
    """Calculates monthly seasonal standard deviation of the timeseries data in a DataFrame

    Parameters
    ----------

    merged_data: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        A pandas dataframe with a string type index of date representations and the monthly seasonal standard deviation
        as float values in the columns.

    Examples
    --------

    >>> import hydrostats.data as hd
    >>> import pandas as pd
    >>> pd.options.display.max_rows = 15

    The data URLs contain streamflow data from two different models, and are provided from the Hydrostats Github page

    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('Streamflow Prediction Tool', 'GLOFAS'))

    >>> hd.monthly_std_dev(merged_df)
        Streamflow Prediction Tool       GLOFAS
    01                 2483.087791  2346.587412
    02                 2049.872689  1834.931830
    03                 2762.489873  2401.929369
    04                 4251.358318  4001.410267
    05                 5458.296296  4597.889041
    06                 4858.578846  4410.784791
    07                 4494.550854  3802.392524
    08                 4261.406429  3626.662349
    09                 4645.650733  4011.522281
    10                 4410.965472  3755.362766
    11                 4760.001179  3773.190543
    12                 4449.865762  3885.481991

    """
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m"))
    return a.std()


def remove_nan_df(merged_dataframe):
    """Drops rows with NaN, zero, negative, and inf values from a pandas dataframe

    Parameters
    ----------

    merged_dataframe: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    Returns
    -------
    DataFrame
        Pandas dataframe with rows containing NaN, zero, negative, and inf values removed.

    Examples
    --------

    >>> import pandas as pd
    >>> import numpy as np
    >>> import hydrostats.data as hd

    An example dataframe is created with invalid values inserted into it.

    >>> data = np.random.rand(15, 2)
    >>> data[0, 0] = data[1, 1] = np.nan
    >>> data[2, 0] = data[3, 1] = np.inf
    >>> data[4, 0] = data[5, 1] = 0
    >>> data[6, 0] = data[7, 1] = -0.1
    >>> example_df = pd.DataFrame(data=data, index=pd.date_range('1980-01-01', periods=15))
    >>> example_df
                       0         1
    1980-01-01       NaN  0.358903
    1980-01-02  0.074718       NaN
    1980-01-03       inf  0.515931
    1980-01-04  0.498002       inf
    1980-01-05  0.000000  0.617974
    1980-01-06  0.522522  0.000000
    1980-01-07 -0.100000  0.116625
    1980-01-08  0.588054 -0.100000
    1980-01-09  0.691136  0.561178
    1980-01-10  0.998170  0.432411
    1980-01-11  0.424473  0.599110
    1980-01-12  0.330988  0.469158
    1980-01-13  0.633894  0.191701
    1980-01-14  0.183241  0.784494
    1980-01-15  0.681419  0.303280

    Now the NaN, inf, negative, and zero values can be remove with this function.

    >>> hd.remove_nan_df(example_df)
                       0         1
    1980-01-09  0.691136  0.561178
    1980-01-10  0.998170  0.432411
    1980-01-11  0.424473  0.599110
    1980-01-12  0.330988  0.469158
    1980-01-13  0.633894  0.191701
    1980-01-14  0.183241  0.784494
    1980-01-15  0.681419  0.303280

    """
    # Drops Zeros and negatives
    merged_dataframe = merged_dataframe.loc[~(merged_dataframe <= 0).any(axis=1)]
    # Replaces infinites with nans
    merged_dataframe = merged_dataframe.replace([inf, -inf], nan)
    # Drops Nan
    merged_dataframe = merged_dataframe.dropna()

    return merged_dataframe


def seasonal_period(merged_dataframe, daily_period, time_range=None, numpy=False):
    """Creates a dataframe with a specified seasonal period

    Parameters
    ----------

    merged_dataframe: DataFrame
        A pandas DataFrame with a datetime index and columns containing float type values.

    daily_period: tuple of str
        A list of length two with strings representing the start and end dates of the seasonal period (e.g.
        [01-01, 01-31] for Jan 1 to Jan 31.

    time_range: tuple of str
        A tuple of string values representing the start and end dates of the time range. Format is YYYY-MM-DD.

    numpy: bool
        If True, two numpy arrays will be returned instead of a pandas dataframe

    Returns
    -------
    DataFrame
        Pandas dataframe that has been truncated to fit the parameters specified for the seasonal period.

    Examples
    --------
    >>> import pandas
    >>> pd.options.display.max_rows = 15
    >>> import numpy as np
    >>> import hydrostats.data as hd

    Here an example DataFrame is made with appx three years of data.

    >>> example_df = pd.DataFrame(data=np.random.rand(1000, 2), index=pd.date_range('2000-01-01', periods=1000), columns=['Simulated', 'Observed'])
                Simulated  Observed
    2000-01-01   0.862726  0.056597
    2000-01-02   0.979643  0.915072
    2000-01-03   0.857667  0.965057
    2000-01-04   0.011238  0.033678
    2000-01-05   0.011390  0.401728
    2000-01-06   0.056505  0.047417
    2000-01-07   0.615151  0.134103
                   ...       ...
    2002-09-20   0.883156  0.272355
    2002-09-21   0.595319  0.406609
    2002-09-22   0.415106  0.826873
    2002-09-23   0.399449  0.656040
    2002-09-24   0.243404  0.561899
    2002-09-25   0.879932  0.551347
    2002-09-26   0.787526  0.887288
    [1000 rows x 2 columns]

    Using this function, a new dataframe containing only the data values in january is returned.

    >>> seasonal_df_jan = hd.seasonal_period(example_df, ('01-01', '01-31'))
                Simulated  Observed
    2000-01-01   0.862726  0.056597
    2000-01-02   0.979643  0.915072
    2000-01-03   0.857667  0.965057
    2000-01-04   0.011238  0.033678
    2000-01-05   0.011390  0.401728
    2000-01-06   0.056505  0.047417
    2000-01-07   0.615151  0.134103
                   ...       ...
    2002-01-25   0.230580  0.363213
    2002-01-26   0.579899  0.370847
    2002-01-27   0.317925  0.120410
    2002-01-28   0.196034  0.035715
    2002-01-29   0.245429  0.974162
    2002-01-30   0.156166  0.544797
    2002-01-31   0.158595  0.311630
    [93 rows x 2 columns]

    We can also specify a time range if we only want the months of January in the year 2000 and 2001

    >>> seasonal_df_jan = hd.seasonal_period(example_df, ('01-01', '01-31'), time_range=('2000-01-01', '2001-12-31'))
                Simulated  Observed
    2000-01-01   0.862726  0.056597
    2000-01-02   0.979643  0.915072
    2000-01-03   0.857667  0.965057
    2000-01-04   0.011238  0.033678
    2000-01-05   0.011390  0.401728
    2000-01-06   0.056505  0.047417
    2000-01-07   0.615151  0.134103
                   ...       ...
    2001-01-25   0.119188  0.043076
    2001-01-26   0.896280  0.282883
    2001-01-27   0.659078  0.230265
    2001-01-28   0.667826  0.383687
    2001-01-29   0.298459  0.738100
    2001-01-30   0.336499  0.189036
    2001-01-31   0.571562  0.783718
    [62 rows x 2 columns]

    """
    # Making a copy to avoid changing the original df
    merged_df_copy = merged_dataframe.copy()

    if time_range:
        # Setting the time range
        merged_df_copy = merged_df_copy.loc[time_range[0]: time_range[1]]

    # Setting a placeholder for the datetime string values
    merged_df_copy.insert(loc=0, column='placeholder', value=merged_df_copy.index.strftime('%m-%d'))

    # getting the start and end of the seasonal period
    start = daily_period[0]
    end = daily_period[1]

    # Getting the seasonal period
    if start < end:
        merged_df_copy = merged_df_copy.loc[(merged_df_copy['placeholder'] >= start) &
                                            (merged_df_copy['placeholder'] <= end)]
    else:
        merged_df_copy = merged_df_copy.loc[(merged_df_copy['placeholder'] >= start) |
                                            (merged_df_copy['placeholder'] <= end)]
    # Dropping the placeholder
    merged_df_copy = merged_df_copy.drop(columns=['placeholder'])

    if numpy:
        return merged_df_copy.iloc[:, 0].values, merged_df_copy.iloc[:, 1].values
    else:
        return merged_df_copy


if __name__ == "__main__":
    pass
    # Defining the URLs of the datasets
    # sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena' \
    #            r'-calamar_interim_data.csv '
    # glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena' \
    #              r'-calamar_ECMWF_data.csv '
    # # Merging the data
    # merged_df = merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))
    #
    # merged_df.to_pickle("tests/Files_for_tests/merged_df.pkl", protocol=2)



# python 3.6
# -*- coding: utf-8 -*-
"""
The analyze module contains functions that perform a more complex analysis of simulated and observed
time series data. It allows users to make tables with metrics that they choose as well as different
date ranges. It also allows users to run a time lag analysis of two time series.
"""
from __future__ import division
from hydrostats.metrics import list_of_metrics
from HydroErr.HydroErr import treat_values
import pandas as pd
from hydrostats.data import seasonal_period
import calendar
import numpy as np
import matplotlib.pyplot as plt

__all__ = ['make_table', 'time_lag']


def make_table(merged_dataframe, metrics, seasonal_periods=None, mase_m=1, dmod_j=1,
               nse_mod_j=1, h6_mhe_k=1, h6_ahe_k=1, h6_rmshe_k=1, d1_p_obs_bar_p=None,
               lm_x_obs_bar_p=None, kge2009_s=(1, 1, 1), kge2012_s=(1, 1, 1), replace_nan=None, replace_inf=None,
               remove_neg=False, remove_zero=False, location=None):
    """Create a table of user selected metrics with optional seasonal analysis.

    Creates a table with metrics as specified by the user. Seasonal periods can also be
    specified in order to compare different seasons and how well the simulated data matches the
    observed data. Has options to save the table to either a csv or an excel workbook. Also has
    an option to add a column for the location of the data.

    Parameters
    ----------
    merged_dataframe: DataFrame
        A pandas dataframe that has two columns of predicted data (Col 0) and observed data (Col 1)
        with a datetime index.

    metrics: list of str
        A list of all the metrics that the user wants to calculate. The metrics abbreviations must
        be used and can be found in the table of all the metrics in the documentation.

    seasonal_periods: 2D list of str, optional
        If given, specifies the seasonal periods that the user wants to analyze (e.g. [['06-01',
        '06-30'], ['08-12', '11-23']] would analyze the dates from June 1st to June 30th and also
        August 8th to November 23). Note that the entire time series is analyzed with the selected
        metrics by default.

    mase_m: int, Optional
        Parameter for the mean absolute scaled error (MASE) metric.

    dmod_j: int or float, optional
        Parameter for the modified index of agreement (dmod) metric.

    nse_mod_j: int or float, optional
        Parameter for the modified Nash-Sutcliffe (nse_mod) metric.

    h6_mhe_k: int or float, optional
        Parameter for the H6 (MHE) metric.

    h6_ahe_k: int or float, optional
        Parameter for the H6 (AHE) metric

    h6_rmshe_k: int or float, optional
        Parameter for the H6 (RMSHE) metric

    d1_p_obs_bar_p: float, optional
        Parameter fot the Legate McCabe Index of Agreement (d1_p).

    lm_x_obs_bar_p: float, optional
        Parameter for the Lagate McCabe Efficiency Index (lm_index).

    kge2009_s: tuple of floats
        A tuple of floats of length three signifying how to weight the three values used in the Kling Gupta (2009)
        metric.

    kge2012_s: tuple of floats
        A tuple of floats of length three signifying how to weight the three values used in the Kling Gupta (2012)
        metric.

    replace_nan: float, optional
        If given, indicates which value to replace NaN values with in the two arrays. If None, when
        a NaN value is found at the i-th position in the observed OR simulated array, the i-th value
        of the observed and simulated array are removed before the computation.

    replace_inf: float, optional
        If given, indicates which value to replace Inf values with in the two arrays. If None, when
        an inf value is found at the i-th position in the observed OR simulated array, the i-th
        value of the observed and simulated array are removed before the computation.

    remove_neg: boolean, optional
        If True, when a negative value is found at the i-th position in the observed OR simulated
        array, the i-th value of the observed AND simulated array are removed before the
        computation.

    remove_zero: boolean, optional
        If true, when a zero value is found at the i-th position in the observed OR simulated
        array, the i-th value of the observed AND simulated array are removed before the
        computation.

    location: str
        The name of the location that will be created as a column in the table that is created.
        Useful for creating a large table with different datasets.

    Returns
    -------
    DataFrame
        Dataframe with rows containing the metric values at the different time ranges, and columns
        containing the metrics specified.

    Notes
    -----
    If desired, users can export the tables to a CSV or Excel Workbook. This can be done using the built in methods
    of pandas. A link to CSV method can be found at
    `<https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html>`_ and a link to the Excel
    method can be found at
    `<https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_excel.html>`_

    Examples
    --------
    First we need to get some data. The data here is pulled from the Streamflow Predication Tool
    model and the ECMWF forecasting model. We are comparing the two models in this example.

    >>> import hydrostats.analyze as ha
    >>> import hydrostats.data as hd
    >>>
    >>> # Defining the URLs of the datasets
    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> # Merging the data
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))

    Here we make a table and print the results

    >>> my_metrics = ['MAE', 'r2', 'NSE', 'KGE (2012)']
    >>> seasonal = [['01-01', '03-31'], ['04-01', '06-30'], ['07-01', '09-30'], ['10-01', '12-31']]
    >>> table = ha.make_table(merged_df, my_metrics, seasonal, remove_neg=True, remove_zero=True, location='Magdalena')
    >>> table
                             Location          MAE     ...           NSE  KGE (2012)
    Full Time Series        Magdalena  1157.669988     ...      0.873684    0.872871
    January-01:March-31     Magdalena   631.984177     ...      0.861163    0.858187
    April-01:June-30        Magdalena  1394.640050     ...      0.813737    0.876890
    July-01:September-30    Magdalena  1188.542871     ...      0.829492    0.831188
    October-01:December-31  Magdalena  1410.852917     ...      0.793927    0.791257

    [5 rows x 5 columns]

    We can also write the table to a CSV or Excel worksheet.

    >>> ha.make_table(merged_df, my_metrics, seasonal, remove_neg=True, remove_zero=True, location='Magdalena', to_csv='magdalena_table.csv')
                             Location          MAE     ...           NSE  KGE (2012)
    Full Time Series        Magdalena  1157.669988     ...      0.873684    0.872871
    January-01:March-31     Magdalena   631.984177     ...      0.861163    0.858187
    April-01:June-30        Magdalena  1394.640050     ...      0.813737    0.876890
    July-01:September-30    Magdalena  1188.542871     ...      0.829492    0.831188
    October-01:December-31  Magdalena  1410.852917     ...      0.793927    0.791257

    [5 rows x 5 columns]

    """

    # Creating a list for all of the metrics for all of the seasons
    complete_metric_list = []

    # Creating an index list
    index_array = ['Full Time Series']
    if seasonal_periods is not None:
        seasonal_periods_names = []
        for i in seasonal_periods:
            month_1 = calendar.month_name[int(i[0][:2])]
            month_2 = calendar.month_name[int(i[1][:2])]
            name = month_1 + i[0][2:] + ':' + month_2 + i[1][2:]
            seasonal_periods_names.append(name)
        index_array.extend(seasonal_periods_names)

    # Creating arrays for sim and obs with all the values if a merged dataframe is given
    sim_array = merged_dataframe.iloc[:, 0].values
    obs_array = merged_dataframe.iloc[:, 1].values

    # Getting a list of the full time series
    full_time_series_list = list_of_metrics(metrics=metrics, abbr=True, sim_array=sim_array,
                                            obs_array=obs_array, mase_m=mase_m, dmod_j=dmod_j,
                                            nse_mod_j=nse_mod_j, h6_mhe_k=h6_mhe_k,
                                            h6_ahe_k=h6_ahe_k, h6_rmshe_k=h6_rmshe_k,
                                            d1_p_obs_bar_p=d1_p_obs_bar_p, kge2009_s=kge2009_s, kge2012_s=kge2012_s,
                                            lm_x_obs_bar_p=lm_x_obs_bar_p, replace_nan=replace_nan,
                                            replace_inf=replace_inf, remove_neg=remove_neg,
                                            remove_zero=remove_zero)

    # Appending the full time series list to the entire list:
    complete_metric_list.append(full_time_series_list)

    # Calculating metrics for the seasonal periods
    if seasonal_periods is not None:
        for time in seasonal_periods:
            temp_df = seasonal_period(merged_dataframe, time)
            sim_array = temp_df.iloc[:, 0].values
            obs_array = temp_df.iloc[:, 1].values

            seasonal_metric_list = list_of_metrics(
                metrics=metrics, sim_array=sim_array, abbr=True, obs_array=obs_array,
                mase_m=mase_m, dmod_j=dmod_j, nse_mod_j=nse_mod_j, h6_mhe_k=h6_mhe_k,
                h6_ahe_k=h6_ahe_k, h6_rmshe_k=h6_rmshe_k, d1_p_obs_bar_p=d1_p_obs_bar_p,
                lm_x_obs_bar_p=lm_x_obs_bar_p, kge2009_s=kge2009_s, kge2012_s=kge2012_s,
                replace_nan=replace_nan, replace_inf=replace_inf,
                remove_neg=remove_neg, remove_zero=remove_zero
            )

            complete_metric_list.append(seasonal_metric_list)

    table_df_final = pd.DataFrame(complete_metric_list, index=index_array, columns=metrics)

    if location is not None:
        col_values = [location for i in range(table_df_final.shape[0])]
        table_df_final.insert(loc=0, column='Location', value=np.array(col_values))

    return table_df_final


def time_lag(merged_dataframe, metrics, interp_freq='6H', interp_type='pchip',
             shift_range=(-30, 30), mase_m=1, dmod_j=1, nse_mod_j=1, h6_mhe_k=1,
             h6_ahe_k=1, h6_rmshe_k=1, d1_p_obs_bar_p=None, lm_x_obs_bar_p=None, replace_nan=None,
             replace_inf=None, remove_neg=False, remove_zero=False, plot=False,
             plot_title='Metric Values as Different Lags', ylabel='Metric Value',
             xlabel='Number of Lags', save_fig=None, figsize=(10, 6), station=None):
    """Check metric values between simulated and observed data at different time lags.

    Runs a time lag analysis to check for potential timing errors in the simulated data. Can also create a
    plot using matplotlib of the metric values at different shifts.

    Parameters
    ----------
    merged_dataframe: DataFrame
        A pandas dataframe that has two columns of predicted data (Col 0) and observed data (Col 1)
        with a datetime index.

    metrics: list of str
        Metric abbreviations that the user would like to use in their lag analysis.

    interp_freq: str, optional
        Frequency of the interpolation for both the simulated and observed time series.

    interp_type: str, optional
        Type of interpolation. Options are 'linear', 'pchip', or 'cubic'.

    shift_range: tuple of 2 floats, optional
        If given, specifies the range of the lag shifts. For example, if (-10, 10) was given, the
        function would shift the simulated array for all of the range between -10 and 10.

    mase_m: int, Optional
        Parameter for the mean absolute scaled error (MASE) metric.

    dmod_j: int or float, optional
        Parameter for the modified index of agreement (dmod) metric.

    nse_mod_j: int or float, optional
        Parameter for the modified Nash-Sutcliffe (nse_mod) metric.

    h6_mhe_k: int or float, optional
        Parameter for the H6 (MHE) metric.

    h6_ahe_k: int or float, optional
        Parameter for the H6 (AHE) metric

    h6_rmshe_k: int or float, optional
        Parameter for the H6 (RMSHE) metric

    d1_p_obs_bar_p: float, optional
        Parameter fot the Legate McCabe Index of Agreement (d1_p).

    lm_x_obs_bar_p: float, optional
        Parameter for the Lagate McCabe Efficiency Index (lm_index).

    replace_nan: float, optional
        If given, indicates which value to replace NaN values with in the two arrays. If None, when
        a NaN value is found at the i-th position in the observed OR simulated array, the i-th value
        of the observed and simulated array are removed before the computation.

    replace_inf: float, optional
        If given, indicates which value to replace Inf values with in the two arrays. If None, when
        an inf value is found at the i-th position in the observed OR simulated array, the i-th
        value of the observed and simulated array are removed before the computation.

    remove_neg: boolean, optional
        If True, when a negative value is found at the i-th position in the observed OR simulated
        array, the i-th value of the observed AND simulated array are removed before the
        computation.

    remove_zero: boolean, optional
        If true, when a zero value is found at the i-th position in the observed OR simulated
        array, the i-th value of the observed AND simulated array are removed before the
        computation.

    plot: bool, optional
        If True, a plot will be created that visualizes the different error metrics at the
        different time lags.

    plot_title: str, optional
        If the plot parameter is true, this parameter will set the title of the plot.

    ylabel: str, optional
        If the plot parameter is true, this parameter will set the y axis label of the plot.

    xlabel: str, optional
        if the plot parameter is true, this parameter will set the x axis label of the plot.

    save_fig: str, optional
        If given, will save the matplotlib plot to the specified directory with the specified file
        name (e.g. /path/to/plot.png). A list of all available types of figures is given in
        :ref:`plot_types`.

    figsize: tuple of length 2
        Tuple that specifies the horizontal and vertical lengths of the plot, in inches.

    station: str
        The station of analysis. Includes the station in the table if given.

    Notes
    -----
    If desired, users can export the tables to a CSV or Excel Workbook. This can be done using the built in methods
    of pandas. A link to CSV method can be found at
    `<https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html>`_ and a link to the Excel
    method can be found at
    `<https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_excel.html>`_

    Returns
    -------
    Tuple of DataFrames
        The first DataFrame contains all of the metric values at different time lags, while the 
        second dataframe contains the maximum and minimum metric values throughout the time lag, 
        and the index of the maximum and minimum time lag values.

    Examples
    --------

    Using data from the Streamflow prediction tool RAPID model and the ECMWF model, we can conpare
    the two at different time lags

    >>> import hydrostats.analyze as ha
    >>> import hydrostats.data as hd
    >>> pd.options.display.max_columns = 50
    >>>
    >>> # Defining the URLs of the datasets
    >>> sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    >>> glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'
    >>> # Merging the data
    >>> merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=('SFPT', 'GLOFAS'))

    There are two dataframes that are returned as part of the analysis.

    >>> # Running the lag analysis
    >>> time_lag_df, summary_df = ha.time_lag(merged_df, metrics=['ME', 'r2', 'RMSE', 'KGE (2012)', 'NSE'])
    >>> summary_df
                        Max  Max Lag Number          Min  Min Lag Number
    ME           174.740510           -28.0   174.740510           -24.0
    r2             0.891854            -2.0     0.646193           -30.0
    RMSE        3142.795474           -30.0  1754.455598            -2.0
    KGE (2012)     0.877116            -2.0     0.775328           -30.0
    NSE            0.856358            -2.0     0.539076           -30.0
    
    A plot can be created that visualizes the different metrics throughout the time lags. It can be
    saved using the savefig parameter as well if desired.

    >>> _, _ = ha.time_lag(merged_df, metrics=['r2', 'KGE (2012)', 'dr'], plot=True)

    .. image:: /Figures/lag_plot1.png

    """

    # Making a new time index to be able to interpolate the time series to the required input
    new_index = pd.date_range(merged_dataframe.index[0], merged_dataframe.index[-1],
                              freq=interp_freq)

    # Reindexing the dataframe and interpolating it
    try:
        merged_dataframe = merged_dataframe.reindex(new_index)
        merged_dataframe = merged_dataframe.interpolate(interp_type)
    except Exception as e:
        raise RuntimeError("Error while interpolating, please make sure that you don't have "
                              "duplicate dates in your time series data.")

    # Making arrays to compare the metric value at different time steps
    sim_array = merged_dataframe.iloc[:, 0].values
    obs_array = merged_dataframe.iloc[:, 1].values

    sim_array, obs_array = treat_values(sim_array, obs_array, replace_nan=replace_nan, replace_inf=replace_inf,
                                        remove_zero=remove_zero, remove_neg=remove_neg)

    # Creating a list to append the values of shift to
    shift_list = []

    # Creating a list of all the time shifts specified by the user
    lag_array = np.arange(shift_range[0], shift_range[1] + 1)

    # Looping through the list of lags and appending the metric value to the shift list
    for i in lag_array:
        sim_array_temp = np.roll(sim_array, i)

        lag_metrics = list_of_metrics(
            metrics=metrics, sim_array=sim_array_temp, obs_array=obs_array, mase_m=mase_m,
            abbr=True, dmod_j=dmod_j, nse_mod_j=nse_mod_j, h6_mhe_k=h6_mhe_k, h6_ahe_k=h6_ahe_k,
            h6_rmshe_k=h6_rmshe_k, d1_p_obs_bar_p=d1_p_obs_bar_p, lm_x_obs_bar_p=lm_x_obs_bar_p,
            replace_nan=replace_nan, replace_inf=replace_inf, remove_neg=remove_neg,
            remove_zero=remove_zero
        )
        shift_list.append(lag_metrics)

    final_array = np.array(shift_list)
    if plot:
        plt.figure(figsize=figsize)

        for i, abbr in enumerate(metrics):
            shift_list_temp = final_array[:, i]
            plt.plot(lag_array, shift_list_temp, label=abbr, alpha=0.7)

        plt.title(plot_title, fontsize=18)
        plt.xlabel(xlabel, fontsize=14)
        plt.ylabel(ylabel, fontsize=14)
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)
        plt.legend()
        if save_fig is None:
            plt.show()
            plt.close()
        else:
            plt.savefig(save_fig)
            plt.close()

    max_lag_array = np.max(final_array, 0)
    max_lag_indices = np.argmax(final_array, 0)
    max_lag_locations = lag_array[max_lag_indices]
    min_lag_array = np.min(final_array, 0)
    min_lag_indices = np.argmin(final_array, 0)
    min_lag_locations = lag_array[min_lag_indices]

    data = np.column_stack((max_lag_array, max_lag_locations, min_lag_array, min_lag_locations))

    summary_df = pd.DataFrame(data=data, index=metrics,
                              columns=["Max", "Max Lag Number", "Min", "Min Lag Number"])

    lag_df = pd.DataFrame(final_array, columns=metrics, index=lag_array)
    lag_df.index.name = "Lag Number"

    # Creating a row with the location name specified
    if station is not None:
        col_values = [station for i in range(summary_df.shape[0])]
        summary_df.insert(loc=0, column='Station', value=np.array(col_values))

    return lag_df, summary_df


if __name__ == "__main__":
    pass
